import datetime
import logging
import time
from collections import defaultdict
from typing import TYPE_CHECKING, AbstractSet, Dict, Mapping, Optional, Sequence, Set, Tuple

import dagster._check as check
from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.declarative_automation.automation_condition import AutomationResult
from dagster._core.definitions.declarative_automation.automation_context import AutomationContext
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey

from ..asset_daemon_cursor import AssetDaemonCursor
from ..base_asset_graph import BaseAssetGraph
from ..freshness_based_auto_materialize import get_expected_data_time_for_asset_key
from .legacy.legacy_context import LegacyRuleEvaluationContext

if TYPE_CHECKING:
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer

from dataclasses import dataclass


@dataclass
class AutomationConditionEvaluator:
    def __init__(
        self,
        *,
        asset_graph: BaseAssetGraph,
        asset_keys: AbstractSet[AssetKey],
        asset_graph_view: AssetGraphView,
        logger: logging.Logger,
        cursor: AssetDaemonCursor,
        data_time_resolver: CachingDataTimeResolver,
        respect_materialization_data_versions: bool,
        # Mapping from run tags to values that should be automatically added run emitted by
        # the declarative scheduling system. This ends up getting sources from places such
        # as https://docs.dagster.io/deployment/dagster-instance#auto-materialize
        # Should this be a supported feature in DS?
        auto_materialize_run_tags: Mapping[str, str],
    ):
        self.asset_graph = asset_graph
        self.asset_keys = asset_keys
        self.asset_graph_view = asset_graph_view
        self.logger = logger
        self.cursor = cursor
        self.data_time_resolver = data_time_resolver
        self.respect_materialization_data_versions = respect_materialization_data_versions
        self.auto_materialize_run_tags = auto_materialize_run_tags

        self.current_results_by_key = {}
        self.condition_cursors = []
        self.expected_data_time_mapping = defaultdict()
        self.to_request = set()
        self.num_checked_assets = 0
        self.num_asset_keys = len(asset_keys)

    asset_graph: BaseAssetGraph
    asset_keys: AbstractSet[AssetKey]
    asset_graph_view: AssetGraphView
    current_results_by_key: Dict[AssetKey, AutomationResult]
    expected_data_time_mapping: Dict[AssetKey, Optional[datetime.datetime]]
    to_request: Set[AssetKeyPartitionKey]
    num_checked_assets: int
    num_asset_keys: int
    logger: logging.Logger
    cursor: AssetDaemonCursor
    data_time_resolver: CachingDataTimeResolver
    respect_materialization_data_versions: bool
    auto_materialize_run_tags: Mapping[str, str]

    @property
    def instance_queryer(self) -> "CachingInstanceQueryer":
        return self.asset_graph_view.get_inner_queryer_for_back_compat()

    @property
    def evaluated_asset_keys_and_parents(self) -> AbstractSet[AssetKey]:
        return {
            parent
            for asset_key in self.asset_keys
            for parent in self.asset_graph.get(asset_key).parent_keys
        } | self.asset_keys

    @property
    def asset_records_to_prefetch(self) -> Sequence[AssetKey]:
        return [key for key in self.evaluated_asset_keys_and_parents if self.asset_graph.has(key)]

    def prefetch(self) -> None:
        """Pre-populate the cached values here to avoid situations in which the new latest_storage_id
        value is calculated using information that comes in after the set of asset partitions with
        new parent materializations is calculated, as this can result in materializations being
        ignored if they happen between the two calculations.
        """
        self.logger.info(
            f"Prefetching asset records for {len(self.asset_records_to_prefetch)} records."
        )
        self.instance_queryer.prefetch_asset_records(self.asset_records_to_prefetch)
        self.logger.info("Done prefetching asset records.")

    def evaluate(self) -> Tuple[Sequence[AutomationResult], AbstractSet[AssetKeyPartitionKey]]:
        self.prefetch()
        for asset_key in self.asset_graph.toposorted_asset_keys:
            # an asset may have already been visited if it was part of a non-subsettable multi-asset
            if asset_key not in self.asset_keys:
                continue

            self.num_checked_assets = self.num_checked_assets + 1
            start_time = time.time()
            self.logger.debug(
                "Evaluating asset"
                f" {asset_key.to_user_string()} ({self.num_checked_assets}/{self.num_asset_keys})"
            )

            try:
                (result, expected_data_time) = self.evaluate_asset(
                    asset_key, self.expected_data_time_mapping, self.current_results_by_key
                )
            except Exception as e:
                raise Exception(
                    f"Error while evaluating conditions for asset {asset_key.to_user_string()}"
                ) from e

            num_requested = result.true_subset.size
            log_fn = self.logger.info if num_requested > 0 else self.logger.debug

            to_request_asset_partitions = result.true_subset.asset_partitions
            to_request_str = ",".join(
                [(ap.partition_key or "No partition") for ap in to_request_asset_partitions]
            )
            self.to_request |= to_request_asset_partitions

            log_fn(
                f"Asset {asset_key.to_user_string()} evaluation result: {num_requested}"
                f" requested ({to_request_str}) ({format(time.time()-start_time, '.3f')} seconds)"
            )

            self.current_results_by_key[asset_key] = result
            self.expected_data_time_mapping[asset_key] = expected_data_time

            # if we need to materialize any partitions of a non-subsettable multi-asset, we need to
            # materialize all of them
            execution_set_keys = self.asset_graph.get(asset_key).execution_set_asset_keys
            if len(execution_set_keys) > 1 and num_requested > 0:
                for neighbor_key in execution_set_keys:
                    self.expected_data_time_mapping[neighbor_key] = expected_data_time

                    # make sure that the true_subset of the neighbor is accurate -- when it was
                    # evaluated it may have had a different requested AssetSubset. however, because
                    # all these neighbors must be executed as a unit, we need to union together
                    # the subset of all required neighbors
                    if neighbor_key in self.current_results_by_key:
                        neighbor_result = self.current_results_by_key[neighbor_key]
                        neighbor_true_subset = result.serializable_evaluation.true_subset._replace(
                            asset_key=neighbor_key
                        )
                        neighbor_evaluation = result.serializable_evaluation._replace(
                            true_subset=neighbor_true_subset
                        )
                        self.current_results_by_key[neighbor_key] = neighbor_result._replace(
                            serializable_evaluation=neighbor_evaluation
                        )
                    self.to_request |= {
                        ap._replace(asset_key=neighbor_key)
                        for ap in result.true_subset.asset_partitions
                    }

        return list(self.current_results_by_key.values()), self.to_request

    def evaluate_asset(
        self,
        asset_key: AssetKey,
        expected_data_time_mapping: Mapping[AssetKey, Optional[datetime.datetime]],
        current_results_by_key: Mapping[AssetKey, AutomationResult],
    ) -> Tuple[AutomationResult, Optional[datetime.datetime]]:
        """Evaluates the auto materialize policy of a given asset key."""
        # convert the legacy AutoMaterializePolicy to an Evaluator
        asset_condition = check.not_none(
            self.asset_graph.get(asset_key).auto_materialize_policy
        ).to_automation_condition()

        legacy_context = LegacyRuleEvaluationContext.create(
            asset_key=asset_key,
            cursor=self.cursor.get_previous_condition_cursor(asset_key),
            condition=asset_condition,
            instance_queryer=self.instance_queryer,
            data_time_resolver=self.data_time_resolver,
            current_results_by_key=current_results_by_key,
            expected_data_time_mapping=expected_data_time_mapping,
            respect_materialization_data_versions=self.respect_materialization_data_versions,
            auto_materialize_run_tags=self.auto_materialize_run_tags,
            logger=self.logger,
        )

        context = AutomationContext.create(
            asset_key=asset_key,
            asset_graph_view=self.asset_graph_view,
            logger=self.logger,
            current_tick_results_by_key=current_results_by_key,
            condition_cursor=self.cursor.get_previous_condition_cursor(asset_key),
            legacy_context=legacy_context,
        )

        result = asset_condition.evaluate(context)
        expected_data_time = get_expected_data_time_for_asset_key(
            legacy_context, will_materialize=result.true_subset.size > 0
        )
        return result, expected_data_time



from dataclasses import dataclass
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    FrozenSet,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Tuple,
    Type,
    TypeVar,
    Union,
)

from dagster._core.asset_graph_view.asset_graph_view import TemporalContext
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.metadata import MetadataMapping, MetadataValue
from dagster._core.definitions.partition import AllPartitionsSubset
from dagster._model.decorator import dagster_model
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils import utc_datetime_from_timestamp

if TYPE_CHECKING:
    from dagster._core.definitions.declarative_automation.automation_condition import (
        AutomationResult,
    )
    from dagster._core.definitions.declarative_automation.automation_context import (
        AutomationContext,
    )

T = TypeVar("T")


@whitelist_for_serdes
@dagster_model(checked=False)
class HistoricalAllPartitionsSubsetSentinel:
    """Serializable indicator that this value was an AllPartitionsSubset at serialization time, but
    the partitions may have changed since that time.
    """


def get_serializable_candidate_subset(
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel],
) -> Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]:
    """Do not serialize the candidate subset directly if it is an AllPartitionsSubset."""
    if isinstance(candidate_subset, AssetSubset) and isinstance(
        candidate_subset.value, AllPartitionsSubset
    ):
        return HistoricalAllPartitionsSubsetSentinel()
    return candidate_subset


@whitelist_for_serdes
class AssetConditionSnapshot(NamedTuple):
    """A serializable snapshot of a node in the AutomationCondition tree."""

    class_name: str
    description: str
    unique_id: str


@whitelist_for_serdes
class AssetSubsetWithMetadata(NamedTuple):
    """An asset subset with metadata that corresponds to it."""

    subset: AssetSubset
    metadata: MetadataMapping

    @property
    def frozen_metadata(self) -> FrozenSet[Tuple[str, MetadataValue]]:
        return frozenset(self.metadata.items())


@whitelist_for_serdes
class AssetConditionEvaluation(NamedTuple):
    """Serializable representation of the results of evaluating a node in the evaluation tree."""

    condition_snapshot: AssetConditionSnapshot
    start_timestamp: Optional[float]
    end_timestamp: Optional[float]

    true_subset: AssetSubset
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata]

    child_evaluations: Sequence["AssetConditionEvaluation"]

    @property
    def asset_key(self) -> AssetKey:
        return self.true_subset.asset_key

    def discarded_subset(self) -> Optional[AssetSubset]:
        """Returns the AssetSubset representing asset partitions that were discarded during this
        evaluation. Note that 'discarding' is a deprecated concept that is only used for backwards
        compatibility.
        """
        if len(self.child_evaluations) != 3:
            return None
        not_discard_evaluation = self.child_evaluations[-1]
        discard_evaluation = not_discard_evaluation.child_evaluations[0]
        return discard_evaluation.true_subset

    def for_child(self, child_unique_id: str) -> Optional["AssetConditionEvaluation"]:
        """Returns the evaluation of a given child condition by finding the child evaluation that
        has an identical hash to the given condition.
        """
        for child_evaluation in self.child_evaluations:
            if child_evaluation.condition_snapshot.unique_id == child_unique_id:
                return child_evaluation

        return None

    def with_run_ids(self, run_ids: AbstractSet[str]) -> "AssetConditionEvaluationWithRunIds":
        return AssetConditionEvaluationWithRunIds(evaluation=self, run_ids=frozenset(run_ids))

    def legacy_num_skipped(self) -> int:
        if len(self.child_evaluations) < 2:
            return 0

        not_skip_evaluation = self.child_evaluations[-1]
        skip_evaluation = not_skip_evaluation.child_evaluations[0]
        return skip_evaluation.true_subset.size - self.legacy_num_discarded()

    def legacy_num_discarded(self) -> int:
        discarded_subset = self.discarded_subset()
        if discarded_subset is None:
            return 0
        return discarded_subset.size


@whitelist_for_serdes
class AssetConditionEvaluationWithRunIds(NamedTuple):
    """A union of an AssetConditionEvaluation and the set of run IDs that have been launched in
    response to it.
    """

    evaluation: AssetConditionEvaluation
    run_ids: FrozenSet[str]

    @property
    def asset_key(self) -> AssetKey:
        return self.evaluation.asset_key

    @property
    def num_requested(self) -> int:
        return self.evaluation.true_subset.size


@whitelist_for_serdes
@dataclass(frozen=True)
class AssetConditionEvaluationState:
    """Incremental state calculated during the evaluation of an AssetCondition. This may be used
    on the subsequent evaluation to make the computation more efficient.

    Attributes:
        previous_evaluation: The computed AssetConditionEvaluation.
        previous_tick_evaluation_timestamp: The evaluation_timestamp at which the evaluation was performed.
        max_storage_id: The maximum storage ID over all events used in this computation.
        extra_state_by_unique_id: A mapping from the unique ID of each condition in the evaluation
            tree to the extra state that was calculated for it, if any.
    """

    previous_evaluation: AssetConditionEvaluation
    previous_tick_evaluation_timestamp: Optional[float]

    max_storage_id: Optional[int]
    extra_state_by_unique_id: Mapping[str, Optional[Union[AssetSubset, Sequence[AssetSubset]]]]

    @property
    def asset_key(self) -> AssetKey:
        return self.previous_evaluation.asset_key

    @property
    def true_subset(self) -> AssetSubset:
        return self.previous_evaluation.true_subset


@whitelist_for_serdes
class AutomationConditionNodeCursor(NamedTuple):
    true_subset: AssetSubset
    candidate_subset: Union[AssetSubset, HistoricalAllPartitionsSubsetSentinel]
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata]
    extra_state: Optional[Union[AssetSubset, Sequence[AssetSubset]]]

    def get_extra_state(self, as_type: Type[T]) -> Optional[T]:
        """Returns the extra_state value if it is of the expected type. Otherwise, returns None."""
        if isinstance(self.extra_state, as_type):
            return self.extra_state
        return None


@whitelist_for_serdes
class AutomationConditionCursor(NamedTuple):
    """Incremental state calculated during the evaluation of a AutomationCondition. This may be used
    on the subsequent evaluation to make the computation more efficient.

    Attributes:
        previous_requested_subset: The subset that was requested for this asset on the previous tick.
        effective_timestamp: The timestamp at which the evaluation was performed.
        last_event_id: The maximum storage ID over all events used in this evaluation.
        node_cursors_by_unique_id: A mapping from the unique ID of each condition in the evaluation
            tree to any incremental state calculated for it.
        result_hash: A unique hash of the result for this tick. Used to determine if anything
            has changed since the last time this was evaluated.
    """

    previous_requested_subset: AssetSubset
    effective_timestamp: float
    last_event_id: Optional[int]

    node_cursors_by_unique_id: Mapping[str, AutomationConditionNodeCursor]
    result_value_hash: str

    @staticmethod
    def backcompat_from_evaluation_state(
        evaluation_state: AssetConditionEvaluationState,
    ) -> "AutomationConditionCursor":
        """Serves as a temporary method to convert from old representation to the new representation."""

        def _get_node_cursors(
            evaluation: AssetConditionEvaluation,
        ) -> Mapping[str, AutomationConditionNodeCursor]:
            node_cursors = {
                evaluation.condition_snapshot.unique_id: AutomationConditionNodeCursor(
                    true_subset=evaluation.true_subset,
                    candidate_subset=evaluation.candidate_subset,
                    subsets_with_metadata=evaluation.subsets_with_metadata,
                    extra_state=evaluation_state.extra_state_by_unique_id.get(
                        evaluation.condition_snapshot.unique_id
                    ),
                )
            }
            for child in evaluation.child_evaluations:
                node_cursors.update(_get_node_cursors(child))

            return node_cursors

        return AutomationConditionCursor(
            previous_requested_subset=evaluation_state.previous_evaluation.true_subset,
            effective_timestamp=evaluation_state.previous_tick_evaluation_timestamp or 0,
            last_event_id=evaluation_state.max_storage_id,
            node_cursors_by_unique_id=_get_node_cursors(evaluation_state.previous_evaluation),
            result_value_hash="",
        )

    @staticmethod
    def from_result(
        context: "AutomationContext", result: "AutomationResult", result_hash: str
    ) -> "AutomationConditionCursor":
        def _gather_node_cursors(r: "AutomationResult"):
            node_cursors = {r.condition_unique_id: r.node_cursor} if r.node_cursor else {}
            for rr in r.child_results:
                node_cursors.update(_gather_node_cursors(rr))
            return node_cursors

        return AutomationConditionCursor(
            previous_requested_subset=result.true_subset,
            effective_timestamp=context.effective_dt.timestamp(),
            last_event_id=context.new_max_storage_id,
            node_cursors_by_unique_id=_gather_node_cursors(result),
            result_value_hash=result_hash,
        )

    @property
    def asset_key(self) -> AssetKey:
        return self.previous_requested_subset.asset_key

    @property
    def temporal_context(self) -> TemporalContext:
        return TemporalContext(
            effective_dt=utc_datetime_from_timestamp(self.effective_timestamp),
            last_event_id=self.last_event_id,
        )



import datetime
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Any, Mapping, NamedTuple, Optional, Sequence, Union

import pendulum

from dagster._annotations import experimental
from dagster._core.asset_graph_view.asset_graph_view import AssetSlice, TemporalContext
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AssetConditionEvaluation,
    AssetConditionSnapshot,
    AssetSubsetWithMetadata,
    AutomationConditionCursor,
    AutomationConditionNodeCursor,
    get_serializable_candidate_subset,
)
from dagster._core.definitions.partition import AllPartitionsSubset
from dagster._core.definitions.time_window_partitions import BaseTimeWindowPartitionsSubset
from dagster._model import DagsterModel
from dagster._utils.security import non_secure_md5_hash_str

if TYPE_CHECKING:
    from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy

    from .automation_context import AutomationContext
    from .operands import (
        CronTickPassedCondition,
        FailedAutomationCondition,
        InLatestTimeWindowCondition,
        InProgressAutomationCondition,
        MissingAutomationCondition,
        NewlyRequestedCondition,
        NewlyUpdatedCondition,
        WillBeRequestedCondition,
    )
    from .operators import (
        AllDepsCondition,
        AndAssetCondition,
        AnyDepsCondition,
        NewlyTrueCondition,
        NotAssetCondition,
        OrAssetCondition,
        SinceCondition,
    )


@experimental
class AutomationCondition(ABC, DagsterModel):
    @property
    def requires_cursor(self) -> bool:
        return False

    @property
    def children(self) -> Sequence["AutomationCondition"]:
        return []

    @property
    @abstractmethod
    def description(self) -> str:
        raise NotImplementedError()

    def get_snapshot(self, unique_id: str) -> AssetConditionSnapshot:
        """Returns a snapshot of this condition that can be used for serialization."""
        return AssetConditionSnapshot(
            class_name=self.__class__.__name__,
            description=self.description,
            unique_id=unique_id,
        )

    def get_unique_id(self, *, parent_unique_id: Optional[str], index: Optional[int]) -> str:
        """Returns a unique identifier for this condition within the broader condition tree."""
        parts = [str(parent_unique_id), str(index), self.__class__.__name__, self.description]
        return non_secure_md5_hash_str("".join(parts).encode())

    def as_auto_materialize_policy(self) -> "AutoMaterializePolicy":
        """Returns an AutoMaterializePolicy which contains this condition."""
        from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy

        return AutoMaterializePolicy.from_automation_condition(self)

    @abstractmethod
    def evaluate(self, context: "AutomationContext") -> "AutomationResult":
        raise NotImplementedError()

    def __and__(self, other: "AutomationCondition") -> "AndAssetCondition":
        from .operators import AndAssetCondition

        # group AndAssetConditions together
        if isinstance(self, AndAssetCondition):
            return AndAssetCondition(operands=[*self.operands, other])
        return AndAssetCondition(operands=[self, other])

    def __or__(self, other: "AutomationCondition") -> "OrAssetCondition":
        from .operators import OrAssetCondition

        # group OrAssetConditions together
        if isinstance(self, OrAssetCondition):
            return OrAssetCondition(operands=[*self.operands, other])
        return OrAssetCondition(operands=[self, other])

    def __invert__(self) -> "NotAssetCondition":
        from .operators import NotAssetCondition

        return NotAssetCondition(operand=self)

    def since(self, reset_condition: "AutomationCondition") -> "SinceCondition":
        """Returns a AutomationCondition that is true if this condition has become true since the
        last time the reference condition became true.
        """
        from .operators import SinceCondition

        return SinceCondition(trigger_condition=self, reset_condition=reset_condition)

    def newly_true(self) -> "NewlyTrueCondition":
        """Returns a AutomationCondition that is true only on the tick that this condition goes
        from false to true for a given asset partition.
        """
        from .operators import NewlyTrueCondition

        return NewlyTrueCondition(operand=self)

    @staticmethod
    def any_deps_match(condition: "AutomationCondition") -> "AnyDepsCondition":
        """Returns a AutomationCondition that is true for an asset partition if at least one partition
        of any of its dependencies evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this asset's dependencies.
        """
        from .operators import AnyDepsCondition

        return AnyDepsCondition(operand=condition)

    @staticmethod
    def all_deps_match(condition: "AutomationCondition") -> "AllDepsCondition":
        """Returns a AutomationCondition that is true for an asset partition if at least one partition
        of all of its dependencies evaluate to True for the given condition.

        Args:
            condition (AutomationCondition): The AutomationCondition that will be evaluated against
                this asset's dependencies.
        """
        from .operators import AllDepsCondition

        return AllDepsCondition(operand=condition)

    @staticmethod
    def missing() -> "MissingAutomationCondition":
        """Returns a AutomationCondition that is true for an asset partition if it has never been
        materialized or observed.
        """
        from .operands import MissingAutomationCondition

        return MissingAutomationCondition()

    @staticmethod
    def in_progress() -> "InProgressAutomationCondition":
        """Returns a AutomationCondition that is true for an asset partition if it is part of an in-progress run."""
        from .operands import InProgressAutomationCondition

        return InProgressAutomationCondition()

    @staticmethod
    def failed() -> "FailedAutomationCondition":
        """Returns a AutomationCondition that is true for an asset partition if its latest run failed."""
        from .operands import FailedAutomationCondition

        return FailedAutomationCondition()

    @staticmethod
    def in_latest_time_window(
        lookback_delta: Optional[datetime.timedelta] = None,
    ) -> "InLatestTimeWindowCondition":
        """Returns a AutomationCondition that is true for an asset partition when it is within the latest
        time window.

        Args:
            lookback_delta (Optional, datetime.timedelta): If provided, the condition will
                return all partitions within the provided delta of the end of the latest time window.
                For example, if this is used on a daily-partitioned asset with a lookback_delta of
                48 hours, this will return the latest two partitions.
        """
        from .operands import InLatestTimeWindowCondition

        return InLatestTimeWindowCondition.from_lookback_delta(lookback_delta)

    @staticmethod
    def will_be_requested() -> "WillBeRequestedCondition":
        """Returns a AutomationCondition that is true for an asset partition if it will be requested this tick."""
        from .operands import WillBeRequestedCondition

        return WillBeRequestedCondition()

    @staticmethod
    def newly_updated() -> "NewlyUpdatedCondition":
        """Returns a AutomationCondition that is true for an asset partition if it has been updated since the previous tick."""
        from .operands import NewlyUpdatedCondition

        return NewlyUpdatedCondition()

    @staticmethod
    def newly_requested() -> "NewlyRequestedCondition":
        """Returns a AutomationCondition that is true for an asset partition if it was requested on the previous tick."""
        from .operands import NewlyRequestedCondition

        return NewlyRequestedCondition()

    @staticmethod
    def cron_tick_passed(
        cron_schedule: str, cron_timezone: str = "UTC"
    ) -> "CronTickPassedCondition":
        """Returns a AutomationCondition that is true for all asset partitions whenever a cron tick of the provided schedule is passed."""
        from .operands import CronTickPassedCondition

        return CronTickPassedCondition(cron_schedule=cron_schedule, cron_timezone=cron_timezone)

    @staticmethod
    def eager() -> "AutomationCondition":
        """Returns a condition which will "eagerly" fill in missing partitions as they are created,
        and ensures unpartitioned assets are updated whenever their dependencies are updated (either
        via scheduled execution or ad-hoc runs).

        Specifically, this is a composite AutomationCondition which is true for an asset partition
        if all of the following are true:

        - The asset partition is within the latest time window
        - At least one of its parents has been updated more recently than it has been requested, or
            the asset partition has never been materialized
        - None of its parent partitions are missing
        - None of its parent partitions are currently part of an in-progress run
        """
        became_missing_or_any_deps_updated = (
            AutomationCondition.missing().newly_true()
            | AutomationCondition.any_deps_match(
                AutomationCondition.newly_updated() | AutomationCondition.will_be_requested()
            )
        )

        any_parent_missing = AutomationCondition.any_deps_match(
            AutomationCondition.missing() & ~AutomationCondition.will_be_requested()
        )
        any_parent_in_progress = AutomationCondition.any_deps_match(
            AutomationCondition.in_progress()
        )
        return (
            AutomationCondition.in_latest_time_window()
            & became_missing_or_any_deps_updated.since(
                AutomationCondition.newly_requested() | AutomationCondition.newly_updated()
            )
            & ~any_parent_missing
            & ~any_parent_in_progress
            & ~AutomationCondition.in_progress()
        )

    @staticmethod
    def cron(cron_schedule: str, cron_timezone: str = "UTC") -> "AutomationCondition":
        """Returns a condition which will materialize asset partitions within the latest time window
        on a given cron schedule, after their parents have been updated. For example, if the
        cron_schedule is set to "0 0 * * *" (every day at midnight), then this rule will not become
        true on a given day until all of its parents have been updated during that same day.

        Specifically, this is a composite AutomationCondition which is true for an asset partition
        if all of the following are true:

        - The asset partition is within the latest time window
        - All parent asset partitions have been updated since the latest tick of the provided cron
            schedule, or will be requested this tick
        - The asset partition has not been requested since the latest tick of the provided cron schedule
        """
        cron_tick_passed = AutomationCondition.cron_tick_passed(cron_schedule, cron_timezone)
        all_deps_updated_since_cron = AutomationCondition.all_deps_match(
            AutomationCondition.newly_updated().since(cron_tick_passed)
            | AutomationCondition.will_be_requested()
        )
        return (
            AutomationCondition.in_latest_time_window()
            & cron_tick_passed.since(AutomationCondition.newly_requested())
            & all_deps_updated_since_cron
        )


class AutomationResult(NamedTuple):
    condition: AutomationCondition
    condition_unique_id: str
    value_hash: str

    start_timestamp: float
    end_timestamp: float

    temporal_context: TemporalContext

    true_slice: AssetSlice
    candidate_slice: AssetSlice

    child_results: Sequence["AutomationResult"]

    node_cursor: Optional[AutomationConditionNodeCursor]
    serializable_evaluation: AssetConditionEvaluation

    extra_state: Any
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata]

    @property
    def asset_key(self) -> AssetKey:
        return self.true_slice.asset_key

    @property
    def true_subset(self) -> AssetSubset:
        return self.true_slice.convert_to_valid_asset_subset()

    @staticmethod
    def _create(
        context: "AutomationContext",
        true_slice: AssetSlice,
        subsets_with_metadata: Sequence[AssetSubsetWithMetadata],
        extra_state: Optional[Union[AssetSubset, Sequence[AssetSubset]]],
        child_results: Sequence["AutomationResult"],
    ) -> "AutomationResult":
        start_timestamp = context.create_time.timestamp()
        end_timestamp = pendulum.now("UTC").timestamp()

        return AutomationResult(
            condition=context.condition,
            condition_unique_id=context.condition_unique_id,
            value_hash=_compute_value_hash(
                condition_unique_id=context.condition_unique_id,
                condition_description=context.condition.description,
                true_slice=true_slice,
                candidate_slice=context.candidate_slice,
                subsets_with_metadata=subsets_with_metadata,
                child_results=child_results,
            ),
            start_timestamp=start_timestamp,
            end_timestamp=end_timestamp,
            temporal_context=context.new_temporal_context,
            true_slice=true_slice,
            candidate_slice=context.candidate_slice,
            subsets_with_metadata=[],
            child_results=child_results,
            extra_state=None,
            node_cursor=_create_node_cursor(
                true_slice=true_slice,
                candidate_slice=context.candidate_slice,
                subsets_with_metadata=subsets_with_metadata,
                extra_state=extra_state,
            )
            if context.condition.requires_cursor
            else None,
            serializable_evaluation=_create_serializable_evaluation(
                context=context,
                true_slice=true_slice,
                candidate_slice=context.candidate_slice,
                subsets_with_metadata=subsets_with_metadata,
                start_timestamp=start_timestamp,
                end_timestamp=end_timestamp,
                child_results=child_results,
            ),
        )

    @staticmethod
    def create_from_children(
        context: "AutomationContext",
        true_slice: AssetSlice,
        child_results: Sequence["AutomationResult"],
        extra_state: Optional[Union[AssetSubset, Sequence[AssetSubset]]] = None,
    ) -> "AutomationResult":
        """Returns a new AssetConditionEvaluation from the given child results."""
        return AutomationResult._create(
            context=context,
            true_slice=true_slice,
            subsets_with_metadata=[],
            extra_state=extra_state,
            child_results=child_results,
        )

    @staticmethod
    def create(
        context: "AutomationContext",
        true_slice: AssetSlice,
        subsets_with_metadata: Sequence[AssetSubsetWithMetadata] = [],
        extra_state: Optional[Union[AssetSubset, Sequence[AssetSubset]]] = None,
    ) -> "AutomationResult":
        """Returns a new AssetConditionEvaluation from the given parameters."""
        return AutomationResult._create(
            context=context,
            true_slice=true_slice,
            subsets_with_metadata=subsets_with_metadata,
            extra_state=extra_state,
            child_results=[],
        )

    def get_child_node_cursors(self) -> Mapping[str, AutomationConditionNodeCursor]:
        node_cursors = {self.condition_unique_id: self.node_cursor} if self.node_cursor else {}
        for child_result in self.child_results:
            node_cursors.update(child_result.get_child_node_cursors())
        return node_cursors

    def get_new_cursor(self) -> AutomationConditionCursor:
        return AutomationConditionCursor(
            previous_requested_subset=self.serializable_evaluation.true_subset,
            effective_timestamp=self.temporal_context.effective_dt.timestamp(),
            last_event_id=self.temporal_context.last_event_id,
            node_cursors_by_unique_id=self.get_child_node_cursors(),
            result_value_hash=self.value_hash,
        )


def _create_node_cursor(
    true_slice: AssetSlice,
    candidate_slice: AssetSlice,
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata],
    extra_state: Optional[Union[AssetSubset, Sequence[AssetSubset]]],
) -> AutomationConditionNodeCursor:
    return AutomationConditionNodeCursor(
        true_subset=true_slice.convert_to_valid_asset_subset(),
        candidate_subset=get_serializable_candidate_subset(
            candidate_slice.convert_to_valid_asset_subset()
        ),
        subsets_with_metadata=subsets_with_metadata,
        extra_state=extra_state,
    )


def _create_serializable_evaluation(
    context: "AutomationContext",
    true_slice: AssetSlice,
    candidate_slice: AssetSlice,
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata],
    start_timestamp: float,
    end_timestamp: float,
    child_results: Sequence[AutomationResult],
) -> AssetConditionEvaluation:
    return AssetConditionEvaluation(
        condition_snapshot=context.condition.get_snapshot(context.condition_unique_id),
        true_subset=true_slice.convert_to_valid_asset_subset(),
        candidate_subset=get_serializable_candidate_subset(
            candidate_slice.convert_to_valid_asset_subset()
        ),
        subsets_with_metadata=subsets_with_metadata,
        start_timestamp=start_timestamp,
        end_timestamp=end_timestamp,
        child_evaluations=[child_result.serializable_evaluation for child_result in child_results],
    )


def _compute_value_hash(
    condition_unique_id: str,
    condition_description: str,
    true_slice: AssetSlice,
    candidate_slice: AssetSlice,
    subsets_with_metadata: Sequence[AssetSubsetWithMetadata],
    child_results: Sequence[AutomationResult],
) -> str:
    """Computes a unique hash representing the values contained within an evaluation result. This
    string will be identical for results which have identical values, allowing us to detect changes
    without serializing the entire value.
    """
    components: Sequence[str] = [
        condition_unique_id,
        condition_description,
        _compute_subset_value_str(true_slice.convert_to_valid_asset_subset()),
        _compute_subset_value_str(candidate_slice.convert_to_valid_asset_subset()),
        *(_compute_subset_with_metadata_value_str(swm) for swm in subsets_with_metadata),
        *(child_result.value_hash for child_result in child_results),
    ]
    return non_secure_md5_hash_str("".join(components).encode("utf-8"))


def _compute_subset_value_str(subset: AssetSubset) -> str:
    """Computes a unique string representing a given AssetSubsets. This string will be equal for
    equivalent AssetSubsets.
    """
    if isinstance(subset.value, bool):
        return str(subset.value)
    elif isinstance(subset.value, AllPartitionsSubset):
        return AllPartitionsSubset.__name__
    elif isinstance(subset.value, BaseTimeWindowPartitionsSubset):
        return str(
            [
                (tw.start.timestamp(), tw.end.timestamp())
                for tw in sorted(subset.value.included_time_windows)
            ]
        )
    else:
        return str(list(sorted(subset.asset_partitions)))


def _compute_subset_with_metadata_value_str(subset_with_metadata: AssetSubsetWithMetadata):
    return _compute_subset_value_str(subset_with_metadata.subset) + str(
        sorted(subset_with_metadata.frozen_metadata)
    )



from abc import ABC
from functools import partial
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    Mapping,
    NamedTuple,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    overload,
)

from typing_extensions import dataclass_transform

import dagster._check as check
from dagster._check import EvalContext, build_check_call_str

TType = TypeVar("TType", bound=Type)
TVal = TypeVar("TVal")


_MODEL_MARKER_VALUE = object()
_MODEL_MARKER_FIELD = (
    "__checkrepublic__"  # "I do want to release this as checkrepublic one day" - schrockn
)
_CHECKED_NEW = "__checked_new__"
_DEFAULTS_NEW = "__defaults_new__"
_INJECTED_DEFAULT_VALS_LOCAL_VAR = "__dm_defaults__"


def _namedtuple_model_transform(
    cls: TType,
    *,
    checked: bool,
    with_new: bool,
    decorator_frames: int,
) -> TType:
    """Transforms the input class in to one that inherits a generated NamedTuple base class
    and:
        * bans tuple methods that don't make sense for a model object
        * creates a run time checked __new__  (optional).
    """
    field_set = getattr(cls, "__annotations__", {})
    defaults = {name: getattr(cls, name) for name in field_set.keys() if hasattr(cls, name)}

    base = NamedTuple(f"_{cls.__name__}", field_set.items())
    nt_new = base.__new__
    if checked:
        eval_ctx = EvalContext.capture_from_frame(
            1 + decorator_frames,
            # inject default values in to the local namespace for reference in generated __new__
            add_to_local_ns={_INJECTED_DEFAULT_VALS_LOCAL_VAR: defaults},
        )
        jit_checked_new = JitCheckedNew(
            field_set,
            defaults,
            base,
            eval_ctx,
            1 if with_new else 0,
        )
        base.__new__ = jit_checked_new  # type: ignore

    elif defaults:
        # allow arbitrary ordering of default values by generating a kwarg only __new__ impl
        eval_ctx = EvalContext(
            global_ns={},
            # inject default values in to the local namespace for reference in generated __new__
            local_ns={_INJECTED_DEFAULT_VALS_LOCAL_VAR: defaults},
        )
        defaults_new = eval_ctx.compile_fn(
            _build_defaults_new(field_set, defaults),
            _DEFAULTS_NEW,
        )
        base.__new__ = defaults_new

    if with_new and cls.__new__ is object.__new__:
        # verify the alignment since it impacts frame capture
        check.failed(f"Expected __new__ on {cls}, add it or switch from the _with_new decorator.")

    # clear default values
    for name in field_set.keys():
        if hasattr(cls, name):
            delattr(cls, name)

    new_type = type(
        cls.__name__,
        (cls, base),
        {  # these will override an implementation on the class if it exists
            "__iter__": _banned_iter,
            "__getitem__": _banned_idx,
            "__hidden_iter__": base.__iter__,
            _MODEL_MARKER_FIELD: _MODEL_MARKER_VALUE,
            "__annotations__": field_set,
            "__nt_new__": nt_new,
            "__bool__": _true,
        },
    )

    return new_type  # type: ignore


@overload
def dagster_model(
    cls: TType,
) -> TType: ...  # Overload for using decorator with no ().


@overload
def dagster_model(
    *,
    checked: bool = True,
) -> Callable[[TType], TType]: ...  # Overload for using decorator used with args.


@dataclass_transform(
    kw_only_default=True,
    frozen_default=True,
)
def dagster_model(
    cls: Optional[TType] = None,
    *,
    checked: bool = True,
) -> Union[TType, Callable[[TType], TType]]:
    """A class decorator that will create an immutable model class based on the defined fields.

    Args:
        checked: Whether or not to generate runtime type checked construction.
        enable_cached_method: Whether or not to support object instance level caching using @cached_method.
        serdes: whitelist this class for serdes, with the defined options if SerdesOptions used.
    """
    if cls:
        return _namedtuple_model_transform(
            cls,
            checked=checked,
            with_new=False,
            decorator_frames=1,
        )
    else:
        return partial(
            _namedtuple_model_transform,
            checked=checked,
            with_new=False,
            decorator_frames=0,
        )


@overload
def dagster_model_custom(
    cls: TType,
) -> TType: ...  # Overload for using decorator with no ().


@overload
def dagster_model_custom(
    *,
    checked: bool = True,
) -> Callable[[TType], TType]: ...  # Overload for using decorator used with args.


def dagster_model_custom(
    cls: Optional[TType] = None,
    *,
    checked: bool = True,
) -> Union[TType, Callable[[TType], TType]]:
    """Variant of the dagster_model decorator to use to opt out of the dataclass_transform decorator behavior.
    This is often doesn't to be able to override __new__, so the type checker respects your constructor.

    @dagster_model_custom
    class Coerced(IHaveNew):
        name: str

        def __new__(cls, name: Optional[str] = None)
            if not name:
                name = "bob"

            return super().__new__(
                cls,
                name=name,
            )



    It would have been cool if we could do that with an argument and @overload but
    from https://peps.python.org/pep-0681/ " When applied to an overload,
    the dataclass_transform decorator still impacts all usage of the function."
    """
    if cls:
        return _namedtuple_model_transform(
            cls,
            checked=checked,
            with_new=True,
            decorator_frames=1,
        )
    else:
        return partial(
            _namedtuple_model_transform,
            checked=checked,
            with_new=True,
            decorator_frames=0,
        )


class IHaveNew:
    """Marker class to be used when overriding new in @dagster_model_custom classes to prevent
    type errors when calling super().__new__.
    """

    if TYPE_CHECKING:

        def __new__(cls, **kwargs): ...


def is_dagster_model(obj) -> bool:
    """Whether or not this object was produced by a dagster_model decorator."""
    return getattr(obj, _MODEL_MARKER_FIELD, None) == _MODEL_MARKER_VALUE


def has_generated_new(obj) -> bool:
    return obj.__new__.__name__ in (_DEFAULTS_NEW, _CHECKED_NEW)


def as_dict(obj) -> Mapping[str, Any]:
    """Creates a dict representation of a model."""
    if not is_dagster_model(obj):
        raise Exception("Only works for @dagster_model decorated classes")

    return {key: value for key, value in zip(obj._fields, obj.__hidden_iter__())}


def copy(obj: TVal, **kwargs) -> TVal:
    """Create a copy of this dagster_model instance, with new values specified as key word args."""
    return obj.__class__(
        **{
            **as_dict(obj),
            **kwargs,
        }
    )


class LegacyNamedTupleMixin(ABC):
    """Mixin to ease migration by adding NamedTuple utility methods.
    Inherit when converting an existing NamedTuple that has callsites to _replace / _asdict, ie.

    @dagster_model
    def AssetSubset(LegacyNamedTupleMixin):
        asset_key: AssetKey
        value: Union[bool, PartitionsSubset]
    """

    def _replace(self, **kwargs):
        return copy(self, **kwargs)

    def _asdict(self):
        return as_dict(self)


class JitCheckedNew:
    """Object that allows us to just-in-time compile a checked __new__ implementation on first use.
    This has two benefits:
        1. Defer processing ForwardRefs until their definitions are in scope.
        2. Avoid up-front cost for unused objects.
    """

    __name__ = _CHECKED_NEW

    def __init__(
        self,
        field_set: Mapping[str, Type],
        defaults: Mapping[str, Any],
        nt_base: Type,
        eval_ctx: EvalContext,
        new_frames: int,
    ):
        self._field_set = field_set
        self._defaults = defaults
        self._nt_base = nt_base
        self._eval_ctx = eval_ctx
        self._new_frames = new_frames  # how many frames of __new__ there are

    def __call__(self, cls, **kwargs):
        # update the context with callsite locals/globals to resolve
        # ForwardRefs that were unavailable at definition time.
        self._eval_ctx.update_from_frame(1 + self._new_frames)

        # ensure check is in scope
        if "check" not in self._eval_ctx.global_ns:
            self._eval_ctx.global_ns["check"] = check

        # jit that shit
        self._nt_base.__new__ = self._eval_ctx.compile_fn(
            self._build_checked_new_str(),
            _CHECKED_NEW,
        )

        return self._nt_base.__new__(cls, **kwargs)

    def _build_checked_new_str(self) -> str:
        kw_args_str, set_calls_str = build_args_and_assignment_strs(self._field_set, self._defaults)

        check_calls = []
        for name, ttype in self._field_set.items():
            call_str = build_check_call_str(
                ttype=ttype,
                name=name,
                eval_ctx=self._eval_ctx,
            )
            check_calls.append(f"{name}={call_str}")

        check_call_block = ",\n        ".join(check_calls)
        return f"""
def __checked_new__(cls{kw_args_str}):
    {set_calls_str}
    return cls.__nt_new__(
        cls,
        {check_call_block}
    )
"""


def _build_defaults_new(field_set: Mapping[str, Type], defaults: Mapping[str, Any]) -> str:
    """Build a __new__ implementation that handles default values."""
    kw_args_str, set_calls_str = build_args_and_assignment_strs(field_set, defaults)
    assign_str = ",\n        ".join([f"{name}={name}" for name in field_set.keys()])
    return f"""
def __defaults_new__(cls{kw_args_str}):
    {set_calls_str}
    return cls.__nt_new__(
        cls,
        {assign_str}
    )
    """


def build_args_and_assignment_strs(
    field_set: Mapping[str, Type],
    defaults: Mapping[str, Any],
) -> Tuple[str, str]:
    """Utility funciton shared between _defaults_new and _checked_new to create the arguments to
    the function as well as any assignment calls that need to happen.
    """
    kw_args = []
    set_calls = []
    for arg in field_set.keys():
        if arg in defaults:
            default = defaults[arg]
            if default is None:
                kw_args.append(f"{arg} = None")
            # dont share class instance of default empty containers
            elif default == []:
                kw_args.append(f"{arg} = None")
                set_calls.append(f"{arg} = {arg} if {arg} is not None else []")
            elif default == {}:
                kw_args.append(f"{arg} = None")
                set_calls.append(f"{arg} = {arg} if {arg} is not None else {'{}'}")
            # fallback to direct reference if unknown
            else:
                kw_args.append(f"{arg} = {_INJECTED_DEFAULT_VALS_LOCAL_VAR}['{arg}']")
        else:
            kw_args.append(arg)

    kw_args_str = ""
    if kw_args:
        kw_args_str = f", *, {', '.join(kw_args)}"

    set_calls_str = ""
    if set_calls:
        set_calls_str = "\n    ".join(set_calls)

    return kw_args_str, set_calls_str


def _banned_iter(*args, **kwargs):
    raise Exception("Iteration is not allowed on `@dagster_model`s.")


def _banned_idx(*args, **kwargs):
    raise Exception("Index access is not allowed on `@dagster_model`s.")


def _true(_):
    return True



from typing import Optional

from .tag_set import NamespacedTagSet as NamespacedTagSet


class StorageKindTagSet(NamespacedTagSet):
    """Tag entries which describe how an asset is stored.

    Args:
        storage_kind (Optional[str]): The storage type of the asset.
            For example, "snowflake" or "s3".
    """

    storage_kind: Optional[str] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster"



from abc import ABC, abstractmethod
from functools import partial
from typing import TYPE_CHECKING, Dict, Generic, Iterable, Optional, Type, TypeVar

from typing_extensions import Self

import dagster._check as check
from dagster._utils.aiodataloader import DataLoader

TResult = TypeVar("TResult")
TKey = TypeVar("TKey")


if TYPE_CHECKING:
    from .instance import DagsterInstance

"""
Loadable - asyncio driven batching

    When resolving a GraphQL operation, it is common to end up with a pattern where a
    set of objects are resolved, and then a field on those objects requires fetching
    a related object by id. To avoid serially fetching those objects by id, something
    must be done. In this past, this batching was solved by manual bespoke classes.

    Loadable is an abstraction that replaces this explicit batching with an implicit scheme
    leveraging the asyncio event loop.

    Within a LoadingContext, requests for the same Loadable object are routed through a shared
    DataLoader instance. This instance accumulates requests that happen in the same event loop
    pass and batches them together. Any subsequent requests for the same ID in the same
    LoadingContext will be served from an in memory cache. This means that mutations need to
    clear the LoadingContext cache.

    The Loadable interface is two methods, `gen` for fetching a single object by its id
    and `gen_many` for loading N by their ids. The "gen" (for generator)[1] name is chosen
    to stand out and make it clear that the method needs to be awaited, which at this time
    is anomalous in a codebase where async is rarely used.

Additional resources:
* https://xuorig.medium.com/the-graphql-dataloader-pattern-visualized-3064a00f319f
[1] Brought to you by the 201X Facebook codebase
"""


class LoadingContext(ABC):
    """A scoped object in which Loadable objects will be fetched in batches and cached.

    Expected to be implemented by request scoped context objects that have access to the DagsterInstance.
    """

    @property
    @abstractmethod
    def instance(self) -> "DagsterInstance":
        raise NotImplementedError()

    @property
    @abstractmethod
    def loaders(self) -> Dict[Type, DataLoader]:
        raise NotImplementedError()

    def get_loader_for(self, ttype: Type["InstanceLoadableBy"]) -> DataLoader:
        if ttype not in self.loaders:
            if not issubclass(ttype, InstanceLoadableBy):
                check.failed(f"{ttype} is not Loadable")

            self.loaders[ttype] = DataLoader(
                batch_load_fn=partial(
                    ttype._batch_load,  # noqa
                    instance=self.instance,
                ),
            )

        return self.loaders[ttype]


# Expected there may be other "Loadable" base classes based on what is needed to load.


class InstanceLoadableBy(ABC, Generic[TKey]):
    """Make An object Loadable by ID of type TKey using a DagsterInstance."""

    @classmethod
    @abstractmethod
    async def _batch_load(
        cls,
        keys: Iterable[TKey],
        instance: "DagsterInstance",
    ) -> Iterable[Optional[Self]]:
        raise NotImplementedError()

    @classmethod
    async def gen(cls, context: LoadingContext, id: TKey) -> Self:
        """Fetch an object by its id."""
        loader = context.get_loader_for(cls)
        return await loader.load(id)

    @classmethod
    async def gen_many(
        cls, context: LoadingContext, ids: Iterable[TKey]
    ) -> Iterable[Optional[Self]]:
        """Fetch N objects by their id."""
        loader = context.get_loader_for(cls)
        return await loader.load_many(ids)



import datetime
import logging
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Any,
    Dict,
    FrozenSet,
    Iterable,
    List,
    Mapping,
    Optional,
    Sequence,
    Set,
    Tuple,
    cast,
)

import pendulum

import dagster._check as check
from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView, TemporalContext
from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.data_version import CachingStaleStatusResolver
from dagster._core.definitions.declarative_automation.automation_condition import AutomationResult
from dagster._core.definitions.events import AssetKey, AssetKeyPartitionKey
from dagster._core.definitions.run_request import RunRequest
from dagster._core.definitions.time_window_partitions import get_time_partitions_def
from dagster._core.instance import DynamicPartitionsStore

from ... import PartitionKeyRange
from ..storage.tags import ASSET_PARTITION_RANGE_END_TAG, ASSET_PARTITION_RANGE_START_TAG
from .asset_daemon_cursor import AssetDaemonCursor
from .auto_materialize_rule import AutoMaterializeRule
from .backfill_policy import BackfillPolicy, BackfillPolicyType
from .base_asset_graph import BaseAssetGraph
from .declarative_automation.serialized_objects import AssetConditionEvaluation
from .partition import PartitionsDefinition, ScheduleType

if TYPE_CHECKING:
    from dagster._core.instance import DagsterInstance
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer  # expensive import


def get_implicit_auto_materialize_policy(
    asset_key: AssetKey, asset_graph: BaseAssetGraph
) -> Optional[AutoMaterializePolicy]:
    """For backcompat with pre-auto materialize policy graphs, assume a default scope of 1 day."""
    auto_materialize_policy = asset_graph.get(asset_key).auto_materialize_policy
    if auto_materialize_policy is None:
        time_partitions_def = get_time_partitions_def(asset_graph.get(asset_key).partitions_def)
        if time_partitions_def is None:
            max_materializations_per_minute = None
        elif time_partitions_def.schedule_type == ScheduleType.HOURLY:
            max_materializations_per_minute = 24
        else:
            max_materializations_per_minute = 1
        rules = {
            AutoMaterializeRule.materialize_on_missing(),
            AutoMaterializeRule.materialize_on_required_for_freshness(),
            AutoMaterializeRule.skip_on_parent_outdated(),
            AutoMaterializeRule.skip_on_parent_missing(),
            AutoMaterializeRule.skip_on_required_but_nonexistent_parents(),
            AutoMaterializeRule.skip_on_backfill_in_progress(),
        }
        if not bool(asset_graph.get_downstream_freshness_policies(asset_key=asset_key)):
            rules.add(AutoMaterializeRule.materialize_on_parent_updated())
        return AutoMaterializePolicy(
            rules=rules,
            max_materializations_per_minute=max_materializations_per_minute,
        )
    return auto_materialize_policy


class AssetDaemonContext:
    def __init__(
        self,
        evaluation_id: int,
        instance: "DagsterInstance",
        asset_graph: BaseAssetGraph,
        cursor: AssetDaemonCursor,
        materialize_run_tags: Optional[Mapping[str, str]],
        observe_run_tags: Optional[Mapping[str, str]],
        auto_observe_asset_keys: Optional[AbstractSet[AssetKey]],
        auto_materialize_asset_keys: Optional[AbstractSet[AssetKey]],
        respect_materialization_data_versions: bool,
        logger: logging.Logger,
        evaluation_time: Optional[datetime.datetime] = None,
    ):
        from dagster._utils.caching_instance_queryer import CachingInstanceQueryer

        self._evaluation_id = evaluation_id
        self._instance_queryer = CachingInstanceQueryer(
            instance, asset_graph, evaluation_time=evaluation_time, logger=logger
        )
        self._data_time_resolver = CachingDataTimeResolver(self.instance_queryer)

        stale_resolver = CachingStaleStatusResolver(
            instance=instance, asset_graph=asset_graph, instance_queryer=self.instance_queryer
        )
        self._asset_graph_view = AssetGraphView(
            temporal_context=TemporalContext(
                effective_dt=self.instance_queryer.evaluation_time,
                last_event_id=instance.event_log_storage.get_maximum_record_id(),
            ),
            stale_resolver=stale_resolver,
        )
        self._data_time_resolver = CachingDataTimeResolver(self.instance_queryer)
        self._cursor = cursor
        self._auto_materialize_asset_keys = auto_materialize_asset_keys or set()
        self._materialize_run_tags = materialize_run_tags
        self._observe_run_tags = observe_run_tags
        self._auto_observe_asset_keys = auto_observe_asset_keys or set()
        self._respect_materialization_data_versions = respect_materialization_data_versions
        self._logger = logger

    @property
    def logger(self) -> logging.Logger:
        return self._logger

    @property
    def instance_queryer(self) -> "CachingInstanceQueryer":
        return self._instance_queryer

    @property
    def data_time_resolver(self) -> CachingDataTimeResolver:
        return self._data_time_resolver

    @property
    def asset_graph_view(self) -> AssetGraphView:
        return self._asset_graph_view

    @property
    def cursor(self) -> AssetDaemonCursor:
        return self._cursor

    @property
    def asset_graph(self) -> BaseAssetGraph:
        return self.instance_queryer.asset_graph

    @property
    def auto_materialize_asset_keys(self) -> AbstractSet[AssetKey]:
        return self._auto_materialize_asset_keys

    @property
    def respect_materialization_data_versions(self) -> bool:
        return self._respect_materialization_data_versions

    @property
    def auto_materialize_run_tags(self) -> Mapping[str, str]:
        return self._materialize_run_tags or {}

    def get_asset_condition_evaluations(
        self,
    ) -> Tuple[Sequence[AutomationResult], AbstractSet[AssetKeyPartitionKey]]:
        """Returns a mapping from asset key to the AutoMaterializeAssetEvaluation for that key, a
        sequence of new per-asset cursors, and the set of all asset partitions that should be
        materialized or discarded this tick.
        """
        from .declarative_automation.automation_condition_evaluator import (
            AutomationConditionEvaluator,
        )

        evaluator = AutomationConditionEvaluator(
            asset_graph=self.asset_graph,
            asset_keys=self.auto_materialize_asset_keys,
            asset_graph_view=self.asset_graph_view,
            logger=self._logger,
            cursor=self.cursor,
            data_time_resolver=self.data_time_resolver,
            respect_materialization_data_versions=self.respect_materialization_data_versions,
            auto_materialize_run_tags=self.auto_materialize_run_tags,
        )
        return evaluator.evaluate()

    def evaluate(
        self,
    ) -> Tuple[Sequence[RunRequest], AssetDaemonCursor, Sequence[AssetConditionEvaluation]]:
        observe_request_timestamp = pendulum.now().timestamp()
        auto_observe_run_requests = (
            get_auto_observe_run_requests(
                asset_graph=self.asset_graph,
                last_observe_request_timestamp_by_asset_key=self.cursor.last_observe_request_timestamp_by_asset_key,
                current_timestamp=observe_request_timestamp,
                run_tags=self._observe_run_tags,
                auto_observe_asset_keys=self._auto_observe_asset_keys,
            )
            if self._auto_observe_asset_keys
            else []
        )

        results, to_request = self.get_asset_condition_evaluations()

        run_requests = [
            *build_run_requests(
                asset_partitions=to_request,
                asset_graph=self.asset_graph,
                run_tags=self.auto_materialize_run_tags,
            ),
            *auto_observe_run_requests,
        ]

        # only record evaluation results where something changed
        updated_evaluations = []
        for result in results:
            previous_cursor = self.cursor.get_previous_condition_cursor(result.asset_key)
            if (
                previous_cursor is None
                or previous_cursor.result_value_hash != result.value_hash
                or not result.true_slice.is_empty
            ):
                updated_evaluations.append(result.serializable_evaluation)

        return (
            run_requests,
            self.cursor.with_updates(
                evaluation_id=self._evaluation_id,
                condition_cursors=[result.get_new_cursor() for result in results],
                newly_observe_requested_asset_keys=[
                    asset_key
                    for run_request in auto_observe_run_requests
                    for asset_key in cast(Sequence[AssetKey], run_request.asset_selection)
                ],
                evaluation_timestamp=self.instance_queryer.evaluation_time.timestamp(),
            ),
            updated_evaluations,
        )


def build_run_requests(
    asset_partitions: Iterable[AssetKeyPartitionKey],
    asset_graph: BaseAssetGraph,
    run_tags: Optional[Mapping[str, str]],
) -> Sequence[RunRequest]:
    assets_to_reconcile_by_partitions_def_partition_key: Mapping[
        Tuple[Optional[PartitionsDefinition], Optional[str]], Set[AssetKey]
    ] = defaultdict(set)

    for asset_partition in asset_partitions:
        assets_to_reconcile_by_partitions_def_partition_key[
            asset_graph.get(asset_partition.asset_key).partitions_def, asset_partition.partition_key
        ].add(asset_partition.asset_key)

    run_requests = []

    for (
        partitions_def,
        partition_key,
    ), asset_keys in assets_to_reconcile_by_partitions_def_partition_key.items():
        tags = {**(run_tags or {})}
        if partition_key is not None:
            if partitions_def is None:
                check.failed("Partition key provided for unpartitioned asset")
            tags.update({**partitions_def.get_tags_for_partition_key(partition_key)})

        for asset_keys_in_repo in asset_graph.split_asset_keys_by_repository(asset_keys):
            run_requests.append(
                # Do not call run_request.with_resolved_tags_and_config as the partition key is
                # valid and there is no config.
                # Calling with_resolved_tags_and_config is costly in asset reconciliation as it
                # checks for valid partition keys.
                RunRequest(
                    asset_selection=list(asset_keys_in_repo),
                    partition_key=partition_key,
                    tags=tags,
                )
            )

    # We don't make public guarantees about sort order, but make an effort to provide a consistent
    # ordering that puts earlier time partitions before later time partitions. Note that, with dates
    # formatted like 12/7/2023, the run requests won't end up in time order. Sorting by converting
    # to time windows seemed more risky from a perf perspective, so we didn't include it here, but
    # it could make sense to actually benchmark that in the future.
    return sorted(run_requests, key=lambda x: x.partition_key if x.partition_key else "")


def build_run_requests_with_backfill_policies(
    asset_partitions: Iterable[AssetKeyPartitionKey],
    asset_graph: BaseAssetGraph,
    run_tags: Optional[Mapping[str, str]],
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Sequence[RunRequest]:
    """If all assets have backfill policies, we should respect them and materialize them according
    to their backfill policies.
    """
    run_requests = []

    asset_partition_keys: Mapping[AssetKey, Set[str]] = {
        asset_key_partition.asset_key: set() for asset_key_partition in asset_partitions
    }
    for asset_partition in asset_partitions:
        if asset_partition.partition_key:
            asset_partition_keys[asset_partition.asset_key].add(asset_partition.partition_key)

    assets_to_reconcile_by_partitions_def_partition_keys_backfill_policy: Mapping[
        Tuple[Optional[PartitionsDefinition], Optional[FrozenSet[str]], Optional[BackfillPolicy]],
        Set[AssetKey],
    ] = defaultdict(set)

    # here we are grouping assets by their partitions def, selected partition keys, and backfill policy.
    for asset_key, partition_keys in asset_partition_keys.items():
        assets_to_reconcile_by_partitions_def_partition_keys_backfill_policy[
            asset_graph.get(asset_key).partitions_def,
            frozenset(partition_keys) if partition_keys else None,
            asset_graph.get(asset_key).backfill_policy,
        ].add(asset_key)

    for (
        partitions_def,
        partition_keys,
        backfill_policy,
    ), asset_keys in assets_to_reconcile_by_partitions_def_partition_keys_backfill_policy.items():
        tags = {**(run_tags or {})}
        if partitions_def is None and partition_keys is not None:
            check.failed("Partition key provided for unpartitioned asset")
        if partitions_def is not None and partition_keys is None:
            check.failed("Partition key missing for partitioned asset")
        if partitions_def is None and partition_keys is None:
            # non partitioned assets will be backfilled in a single run
            run_requests.append(RunRequest(asset_selection=list(asset_keys), tags=tags))
        else:
            run_requests.extend(
                _build_run_requests_with_backfill_policy(
                    list(asset_keys),
                    check.not_none(backfill_policy),
                    check.not_none(partition_keys),
                    check.not_none(partitions_def),
                    tags,
                    dynamic_partitions_store=dynamic_partitions_store,
                )
            )
    return run_requests


def _build_run_requests_with_backfill_policy(
    asset_keys: Sequence[AssetKey],
    backfill_policy: BackfillPolicy,
    partition_keys: FrozenSet[str],
    partitions_def: PartitionsDefinition,
    tags: Dict[str, Any],
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Sequence[RunRequest]:
    run_requests = []
    partition_subset = partitions_def.subset_with_partition_keys(partition_keys)
    partition_key_ranges = partition_subset.get_partition_key_ranges(
        partitions_def, dynamic_partitions_store=dynamic_partitions_store
    )
    for partition_key_range in partition_key_ranges:
        # We might resolve more than one partition key range for the given partition keys.
        # We can only apply chunking on individual partition key ranges.
        if backfill_policy.policy_type == BackfillPolicyType.SINGLE_RUN:
            run_requests.append(
                _build_run_request_for_partition_key_range(
                    asset_keys=list(asset_keys),
                    partition_range_start=partition_key_range.start,
                    partition_range_end=partition_key_range.end,
                    run_tags=tags,
                )
            )
        else:
            run_requests.extend(
                _build_run_requests_for_partition_key_range(
                    asset_keys=list(asset_keys),
                    partitions_def=partitions_def,
                    partition_key_range=partition_key_range,
                    max_partitions_per_run=check.int_param(
                        backfill_policy.max_partitions_per_run, "max_partitions_per_run"
                    ),
                    run_tags=tags,
                    dynamic_partitions_store=dynamic_partitions_store,
                )
            )
    return run_requests


def _build_run_requests_for_partition_key_range(
    asset_keys: Sequence[AssetKey],
    partitions_def: PartitionsDefinition,
    partition_key_range: PartitionKeyRange,
    max_partitions_per_run: int,
    run_tags: Dict[str, str],
    dynamic_partitions_store: DynamicPartitionsStore,
) -> Sequence[RunRequest]:
    """Builds multiple run requests for the given partition key range. Each run request will have at most
    max_partitions_per_run partitions.
    """
    partition_keys = partitions_def.get_partition_keys_in_range(
        partition_key_range, dynamic_partitions_store=dynamic_partitions_store
    )
    partition_range_start_index = partition_keys.index(partition_key_range.start)
    partition_range_end_index = partition_keys.index(partition_key_range.end)

    partition_chunk_start_index = partition_range_start_index
    run_requests = []
    while partition_chunk_start_index <= partition_range_end_index:
        partition_chunk_end_index = partition_chunk_start_index + max_partitions_per_run - 1
        if partition_chunk_end_index > partition_range_end_index:
            partition_chunk_end_index = partition_range_end_index
        partition_chunk_start_key = partition_keys[partition_chunk_start_index]
        partition_chunk_end_key = partition_keys[partition_chunk_end_index]
        run_requests.append(
            _build_run_request_for_partition_key_range(
                asset_keys, partition_chunk_start_key, partition_chunk_end_key, run_tags
            )
        )
        partition_chunk_start_index = partition_chunk_end_index + 1
    return run_requests


def _build_run_request_for_partition_key_range(
    asset_keys: Sequence[AssetKey],
    partition_range_start: str,
    partition_range_end: str,
    run_tags: Dict[str, str],
) -> RunRequest:
    """Builds a single run request for the given asset key and partition key range."""
    tags = {
        **(run_tags or {}),
        ASSET_PARTITION_RANGE_START_TAG: partition_range_start,
        ASSET_PARTITION_RANGE_END_TAG: partition_range_end,
    }
    partition_key = partition_range_start if partition_range_start == partition_range_end else None
    return RunRequest(asset_selection=asset_keys, partition_key=partition_key, tags=tags)


def get_auto_observe_run_requests(
    last_observe_request_timestamp_by_asset_key: Mapping[AssetKey, float],
    current_timestamp: float,
    asset_graph: BaseAssetGraph,
    run_tags: Optional[Mapping[str, str]],
    auto_observe_asset_keys: AbstractSet[AssetKey],
) -> Sequence[RunRequest]:
    assets_to_auto_observe: Set[AssetKey] = set()
    for asset_key in auto_observe_asset_keys:
        last_observe_request_timestamp = last_observe_request_timestamp_by_asset_key.get(asset_key)
        auto_observe_interval_minutes = asset_graph.get(asset_key).auto_observe_interval_minutes

        if auto_observe_interval_minutes and (
            last_observe_request_timestamp is None
            or (
                last_observe_request_timestamp + auto_observe_interval_minutes * 60
                < current_timestamp
            )
        ):
            assets_to_auto_observe.add(asset_key)

    # create groups of asset keys that share the same repository AND the same partitions definition
    partitions_def_and_asset_key_groups: List[Sequence[AssetKey]] = []
    for repository_asset_keys in asset_graph.split_asset_keys_by_repository(assets_to_auto_observe):
        asset_keys_by_partitions_def = defaultdict(list)
        for asset_key in repository_asset_keys:
            partitions_def = asset_graph.get(asset_key).partitions_def
            asset_keys_by_partitions_def[partitions_def].append(asset_key)
        partitions_def_and_asset_key_groups.extend(asset_keys_by_partitions_def.values())

    return [
        RunRequest(asset_selection=list(asset_keys), tags=run_tags)
        for asset_keys in partitions_def_and_asset_key_groups
        if len(asset_keys) > 0
    ]



import datetime
from abc import abstractmethod
from typing import Optional

from dagster._core.asset_graph_view.asset_graph_view import AssetSlice
from dagster._core.definitions.declarative_automation.utils import SerializableTimeDelta
from dagster._serdes.serdes import whitelist_for_serdes
from dagster._utils.schedules import reverse_cron_string_iterator

from ..automation_condition import AutomationCondition, AutomationResult
from ..automation_context import AutomationContext


class SliceAutomationCondition(AutomationCondition):
    """Base class for simple conditions which compute a simple slice of the asset graph."""

    @abstractmethod
    def compute_slice(self, context: AutomationContext) -> AssetSlice: ...

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        # don't compute anything if there are no candidates
        if context.candidate_slice.is_empty:
            true_slice = context.asset_graph_view.create_empty_slice(asset_key=context.asset_key)
        else:
            true_slice = self.compute_slice(context)

        return AutomationResult.create(context, true_slice)


@whitelist_for_serdes
class MissingAutomationCondition(SliceAutomationCondition):
    @property
    def description(self) -> str:
        return "Missing"

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        return context.asset_graph_view.compute_missing_subslice(
            context.asset_key, from_slice=context.candidate_slice
        )


@whitelist_for_serdes
class InProgressAutomationCondition(SliceAutomationCondition):
    @property
    def description(self) -> str:
        return "Part of an in-progress run"

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        return context.asset_graph_view.compute_in_progress_asset_slice(asset_key=context.asset_key)


@whitelist_for_serdes
class FailedAutomationCondition(SliceAutomationCondition):
    @property
    def description(self) -> str:
        return "Latest run failed"

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        return context.asset_graph_view.compute_failed_asset_slice(asset_key=context.asset_key)


@whitelist_for_serdes
class WillBeRequestedCondition(SliceAutomationCondition):
    @property
    def description(self) -> str:
        return "Will be requested this tick"

    def _executable_with_root_context_key(self, context: AutomationContext) -> bool:
        # TODO: once we can launch backfills via the asset daemon, this can be removed
        from dagster._core.definitions.asset_graph import materializable_in_same_run

        root_key = context.root_context.asset_key
        return materializable_in_same_run(
            asset_graph=context.asset_graph_view.asset_graph,
            child_key=root_key,
            parent_key=context.asset_key,
        )

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        current_result = context.current_tick_results_by_key.get(context.asset_key)
        if (
            current_result
            and current_result.true_slice
            and self._executable_with_root_context_key(context)
        ):
            return current_result.true_slice
        else:
            return context.asset_graph_view.create_empty_slice(asset_key=context.asset_key)


@whitelist_for_serdes
class NewlyRequestedCondition(SliceAutomationCondition):
    @property
    def description(self) -> str:
        return "Was requested on the previous tick"

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        return context.previous_requested_slice or context.asset_graph_view.create_empty_slice(
            asset_key=context.asset_key
        )


@whitelist_for_serdes
class NewlyUpdatedCondition(SliceAutomationCondition):
    @property
    def description(self) -> str:
        return "Updated since previous tick"

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        # if it's the first time evaluating, just return the empty slice
        if context.cursor is None:
            return context.asset_graph_view.create_empty_slice(asset_key=context.asset_key)
        else:
            return context.asset_graph_view.compute_updated_since_cursor_slice(
                asset_key=context.asset_key, cursor=context.previous_evaluation_max_storage_id
            )


@whitelist_for_serdes
class CronTickPassedCondition(SliceAutomationCondition):
    cron_schedule: str
    cron_timezone: str

    @property
    def description(self) -> str:
        return f"New tick of {self.cron_schedule} ({self.cron_timezone})"

    def _get_previous_cron_tick(self, effective_dt: datetime.datetime) -> datetime.datetime:
        previous_ticks = reverse_cron_string_iterator(
            end_timestamp=effective_dt.timestamp(),
            cron_string=self.cron_schedule,
            execution_timezone=self.cron_timezone,
        )
        return next(previous_ticks)

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        previous_cron_tick = self._get_previous_cron_tick(context.effective_dt)
        if (
            # no previous evaluation
            context.previous_evaluation_effective_dt is None
            # cron tick was not newly passed
            or previous_cron_tick < context.previous_evaluation_effective_dt
        ):
            return context.asset_graph_view.create_empty_slice(asset_key=context.asset_key)
        else:
            return context.candidate_slice


@whitelist_for_serdes
class InLatestTimeWindowCondition(SliceAutomationCondition):
    serializable_lookback_timedelta: Optional[SerializableTimeDelta] = None

    @staticmethod
    def from_lookback_delta(
        lookback_delta: Optional[datetime.timedelta],
    ) -> "InLatestTimeWindowCondition":
        return InLatestTimeWindowCondition(
            serializable_lookback_timedelta=SerializableTimeDelta.from_timedelta(lookback_delta)
            if lookback_delta
            else None
        )

    @property
    def lookback_timedelta(self) -> Optional[datetime.timedelta]:
        return (
            self.serializable_lookback_timedelta.to_timedelta()
            if self.serializable_lookback_timedelta
            else None
        )

    @property
    def description(self) -> str:
        return (
            f"Within {self.lookback_timedelta} of the end of the latest time window"
            if self.lookback_timedelta
            else "Within latest time window"
        )

    def compute_slice(self, context: AutomationContext) -> AssetSlice:
        return context.asset_graph_view.compute_latest_time_window_slice(
            context.asset_key, lookback_delta=self.lookback_timedelta
        )



import inspect
import os
from pathlib import Path
from typing import TYPE_CHECKING, Any, Callable, List, Optional, Sequence, Union

import dagster._check as check
from dagster._annotations import experimental
from dagster._model import DagsterModel
from dagster._serdes import whitelist_for_serdes

from .metadata_set import (
    NamespacedMetadataSet as NamespacedMetadataSet,
    TableMetadataSet as TableMetadataSet,
)
from .metadata_value import MetadataValue

if TYPE_CHECKING:
    from dagster._core.definitions.assets import AssetsDefinition, SourceAsset
    from dagster._core.definitions.cacheable_assets import CacheableAssetsDefinition

DEFAULT_SOURCE_FILE_KEY = "asset_definition"


@experimental
@whitelist_for_serdes
class LocalFileCodeReference(DagsterModel):
    """Represents a local file source location."""

    file_path: str
    line_number: Optional[int] = None
    label: Optional[str] = None


@experimental
@whitelist_for_serdes
class UrlCodeReference(DagsterModel):
    """Represents a source location which points at a URL, for example
    in source control.
    """

    url: str
    label: Optional[str] = None


@experimental
@whitelist_for_serdes
class CodeReferencesMetadataValue(DagsterModel, MetadataValue["CodeReferencesMetadataValue"]):
    """Metadata value type which represents source locations (locally or otherwise)
    of the asset in question. For example, the file path and line number where the
    asset is defined.

    Attributes:
        sources (List[Union[LocalFileCodeReference, SourceControlCodeReference]]):
            A list of code references for the asset, such as file locations or
            references to source control.
    """

    code_references: List[Union[LocalFileCodeReference, UrlCodeReference]]

    @property
    def value(self) -> "CodeReferencesMetadataValue":
        return self


def local_source_path_from_fn(fn: Callable[..., Any]) -> Optional[LocalFileCodeReference]:
    cwd = os.getcwd()

    origin_file = os.path.abspath(os.path.join(cwd, inspect.getsourcefile(fn)))  # type: ignore
    origin_file = check.not_none(origin_file)
    origin_line = inspect.getsourcelines(fn)[1]

    return LocalFileCodeReference(file_path=origin_file, line_number=origin_line)


class CodeReferencesMetadataSet(NamespacedMetadataSet):
    """Metadata entries that apply to asset definitions and which specify the location where
    source code for the asset can be found.
    """

    code_references: Optional[CodeReferencesMetadataValue] = None

    @classmethod
    def namespace(cls) -> str:
        return "dagster"


def _with_code_source_single_definition(
    assets_def: Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"],
) -> Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"]:
    from dagster._core.definitions.assets import AssetsDefinition

    # SourceAsset doesn't have an op definition to point to - cacheable assets
    # will be supported eventually but are a bit trickier
    if not isinstance(assets_def, AssetsDefinition):
        return assets_def

    metadata_by_key = dict(assets_def.metadata_by_key) or {}

    from dagster._core.definitions.decorators.op_decorator import DecoratedOpFunction

    base_fn = (
        assets_def.op.compute_fn.decorated_fn
        if isinstance(assets_def.op.compute_fn, DecoratedOpFunction)
        else assets_def.op.compute_fn
    )
    source_path = local_source_path_from_fn(base_fn)

    if source_path:
        sources = [source_path]

        for key in assets_def.keys:
            # merge with any existing metadata
            existing_source_code_metadata = CodeReferencesMetadataSet.extract(
                metadata_by_key.get(key, {})
            )
            existing_code_references = (
                existing_source_code_metadata.code_references.code_references
                if existing_source_code_metadata.code_references
                else []
            )
            sources_for_asset: List[Union[LocalFileCodeReference, UrlCodeReference]] = [
                *existing_code_references,
                *sources,
            ]

            metadata_by_key[key] = {
                **metadata_by_key.get(key, {}),
                **CodeReferencesMetadataSet(
                    code_references=CodeReferencesMetadataValue(code_references=sources_for_asset)
                ),
            }

    return assets_def.map_asset_specs(
        lambda spec: spec._replace(metadata=metadata_by_key[spec.key])
    )


def convert_local_path_to_source_control_path(
    base_source_control_url: str,
    repository_root_absolute_path: str,
    local_path: LocalFileCodeReference,
) -> UrlCodeReference:
    source_file_from_repo_root = os.path.relpath(
        local_path.file_path, repository_root_absolute_path
    )
    line_number_suffix = f"#L{local_path.line_number}" if local_path.line_number else ""

    return UrlCodeReference(
        url=f"{base_source_control_url}/{source_file_from_repo_root}{line_number_suffix}",
        label=local_path.label,
    )


def _convert_local_path_to_source_control_path_single_definition(
    base_source_control_url: str,
    repository_root_absolute_path: str,
    assets_def: Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"],
) -> Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"]:
    from dagster._core.definitions.assets import AssetsDefinition

    # SourceAsset doesn't have an op definition to point to - cacheable assets
    # will be supported eventually but are a bit trickier
    if not isinstance(assets_def, AssetsDefinition):
        return assets_def

    metadata_by_key = dict(assets_def.metadata_by_key) or {}

    for key in assets_def.keys:
        existing_source_code_metadata = CodeReferencesMetadataSet.extract(
            metadata_by_key.get(key, {})
        )
        if not existing_source_code_metadata.code_references:
            continue

        sources_for_asset: List[Union[LocalFileCodeReference, UrlCodeReference]] = [
            convert_local_path_to_source_control_path(
                base_source_control_url,
                repository_root_absolute_path,
                source,
            )
            if isinstance(source, LocalFileCodeReference)
            else source
            for source in existing_source_code_metadata.code_references.code_references
        ]

        metadata_by_key[key] = {
            **metadata_by_key.get(key, {}),
            **CodeReferencesMetadataSet(
                code_references=CodeReferencesMetadataValue(code_references=sources_for_asset)
            ),
        }

    return assets_def.map_asset_specs(
        lambda spec: spec._replace(metadata=metadata_by_key[spec.key])
    )


def _build_github_url(url: str, branch: str) -> str:
    return f"{url}/tree/{branch}"


def _build_gitlab_url(url: str, branch: str) -> str:
    return f"{url}/-/tree/{branch}"


@experimental
def link_to_source_control(
    assets_defs: Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"]],
    source_control_url: str,
    source_control_branch: str,
    repository_root_absolute_path: Union[Path, str],
) -> Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"]]:
    """Wrapper function which converts local file path code references to source control URLs
    based on the provided source control URL and branch.

    Args:
        assets_defs (Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]):
            The asset definitions to which source control metadata should be attached.
            Only assets with local file code references (such as those created by
            `with_source_code_references`) will be converted.
        source_control_url (str): The base URL for the source control system. For example,
            "https://github.com/dagster-io/dagster".
        source_control_branch (str): The branch in the source control system, such as "master".
        repository_root_absolute_path (Union[Path, str]): The absolute path to the root of the
            repository on disk. This is used to calculate the relative path to the source file
            from the repository root and append it to the source control URL.
    """
    if "gitlab.com" in source_control_url:
        source_control_url = _build_gitlab_url(source_control_url, source_control_branch)
    elif "github.com" in source_control_url:
        source_control_url = _build_github_url(source_control_url, source_control_branch)
    else:
        raise ValueError(
            "Invalid `source_control_url`."
            " Only GitHub and GitLab are supported for linking to source control at this time."
        )

    return [
        _convert_local_path_to_source_control_path_single_definition(
            base_source_control_url=source_control_url,
            repository_root_absolute_path=str(repository_root_absolute_path),
            assets_def=assets_def,
        )
        for assets_def in assets_defs
    ]


@experimental
def with_source_code_references(
    assets_defs: Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"]],
) -> Sequence[Union["AssetsDefinition", "SourceAsset", "CacheableAssetsDefinition"]]:
    """Wrapper function which attaches local code reference metadata to the provided asset definitions.
    This points to the filepath and line number where the asset body is defined.

    Args:
        assets_defs (Sequence[Union[AssetsDefinition, SourceAsset, CacheableAssetsDefinition]]):
            The asset definitions to which source code metadata should be attached.

    Returns:
        Sequence[AssetsDefinition]: The asset definitions with source code metadata attached.
    """
    return [_with_code_source_single_definition(assets_def) for assets_def in assets_defs]



from .slice_conditions import (
    CronTickPassedCondition as CronTickPassedCondition,
    FailedAutomationCondition as FailedAutomationCondition,
    InLatestTimeWindowCondition as InLatestTimeWindowCondition,
    InProgressAutomationCondition as InProgressAutomationCondition,
    MissingAutomationCondition as MissingAutomationCondition,
    NewlyRequestedCondition as NewlyRequestedCondition,
    NewlyUpdatedCondition as NewlyUpdatedCondition,
    WillBeRequestedCondition as WillBeRequestedCondition,
)



from typing import NamedTuple, Optional

from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView, AssetSlice
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AssetSubsetWithMetadata,
)
from dagster._core.definitions.metadata import MetadataMapping


class AssetSliceWithMetadata(NamedTuple):
    asset_slice: AssetSlice
    metadata: MetadataMapping

    @staticmethod
    def from_asset_subset_with_metadata(
        *, asset_graph_view: AssetGraphView, asset_subset_with_metadata: AssetSubsetWithMetadata
    ) -> Optional["AssetSliceWithMetadata"]:
        metadata = asset_subset_with_metadata.metadata
        if metadata is None:
            return None
        asset_slice = asset_graph_view.get_asset_slice_from_subset(
            asset_subset_with_metadata.subset
        )
        if asset_slice is None:
            return None
        return AssetSliceWithMetadata(asset_slice=asset_slice, metadata=metadata)



from typing import Sequence

from dagster._serdes.serdes import whitelist_for_serdes

from ..automation_condition import AutomationCondition, AutomationResult
from ..automation_context import AutomationContext


@whitelist_for_serdes
class SinceCondition(AutomationCondition):
    trigger_condition: AutomationCondition
    reset_condition: AutomationCondition

    @property
    def requires_cursor(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return (
            "Trigger condition has become true since the last time the reset condition became true."
        )

    @property
    def children(self) -> Sequence[AutomationCondition]:
        return [self.trigger_condition, self.reset_condition]

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        # must evaluate child condition over the entire slice to avoid missing state transitions
        child_candidate_slice = context.asset_graph_view.get_asset_slice(
            asset_key=context.asset_key
        )

        # compute result for trigger condition
        trigger_context = context.for_child_condition(
            self.trigger_condition, child_index=0, candidate_slice=child_candidate_slice
        )
        trigger_result = self.trigger_condition.evaluate(trigger_context)

        # compute result for reset condition
        reset_context = context.for_child_condition(
            self.reset_condition, child_index=1, candidate_slice=child_candidate_slice
        )
        reset_result = self.reset_condition.evaluate(reset_context)

        # take the previous slice that this was true for
        true_slice = context.previous_true_slice or context.asset_graph_view.create_empty_slice(
            asset_key=context.asset_key
        )
        # add in any newly true trigger asset partitions
        true_slice = true_slice.compute_union(trigger_result.true_slice)
        # remove any newly true reset asset partitions
        true_slice = true_slice.compute_difference(reset_result.true_slice)

        return AutomationResult.create_from_children(
            context=context, true_slice=true_slice, child_results=[trigger_result, reset_result]
        )



import datetime
import logging
from typing import TYPE_CHECKING, AbstractSet, Any, Mapping, NamedTuple, Optional

import pendulum

import dagster._check as check
from dagster._core.asset_graph_view.asset_graph_view import (
    AssetGraphView,
    AssetSlice,
    TemporalContext,
)
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_subset import ValidAssetSubset
from dagster._core.definitions.declarative_automation.automation_condition import (
    AutomationCondition,
    AutomationResult,
)
from dagster._core.definitions.declarative_automation.legacy.asset_condition import AssetCondition
from dagster._core.definitions.declarative_automation.serialized_objects import (
    AutomationConditionCursor,
    AutomationConditionNodeCursor,
    HistoricalAllPartitionsSubsetSentinel,
)
from dagster._core.definitions.events import AssetKeyPartitionKey
from dagster._core.definitions.partition import PartitionsDefinition

from .legacy.legacy_context import LegacyRuleEvaluationContext

if TYPE_CHECKING:
    from dagster._core.definitions.base_asset_graph import BaseAssetGraph
    from dagster._utils.caching_instance_queryer import CachingInstanceQueryer


def _has_legacy_condition(condition: AutomationCondition):
    """Detects if the given condition has any legacy rules."""
    if isinstance(condition, AssetCondition):
        return True
    else:
        return any(_has_legacy_condition(child) for child in condition.children)


# This class exists purely for organizational purposes so that we understand
# the interface between automation conditions and the instance much more
# explicitly. This captures all interactions that do not go through AssetGraphView
# so that we do not access the legacy context or the instance queryer directly
# in automation conditions.
class NonAGVInstanceInterface:
    def __init__(self, queryer: "CachingInstanceQueryer"):
        self._queryer = queryer

    def get_asset_subset_updated_after_time(
        self, *, asset_key: AssetKey, after_time: datetime.datetime
    ) -> ValidAssetSubset:
        return self._queryer.get_asset_subset_updated_after_time(
            asset_key=asset_key, after_time=after_time
        )

    def get_parent_asset_partitions_updated_after_child(
        self,
        *,
        asset_partition: AssetKeyPartitionKey,
        parent_asset_partitions: AbstractSet[AssetKeyPartitionKey],
        ignored_parent_keys: AbstractSet[AssetKey],
    ) -> AbstractSet[AssetKeyPartitionKey]:
        return self._queryer.get_parent_asset_partitions_updated_after_child(
            asset_partition=asset_partition,
            parent_asset_partitions=parent_asset_partitions,
            respect_materialization_data_versions=False,
            ignored_parent_keys=ignored_parent_keys,
        )


class AutomationContext(NamedTuple):
    # the slice over which the condition is being evaluated
    candidate_slice: AssetSlice

    # the condition being evaluated
    condition: AutomationCondition
    # a unique identifier for the condition within the broader tree
    condition_unique_id: str

    asset_graph_view: AssetGraphView

    # the context object for the parent condition
    parent_context: Optional["AutomationContext"]

    # the time at which this context object was created
    create_time: datetime.datetime
    logger: logging.Logger

    # a cursor containing information about this asset calculated on the previous tick
    cursor: Optional[AutomationConditionCursor]
    # a mapping of information computed on the current tick for assets which are upstream of this
    # asset
    current_tick_results_by_key: Mapping[AssetKey, AutomationResult]

    non_agv_instance_interface: NonAGVInstanceInterface

    # hack to avoid circular references during pydantic validation
    inner_legacy_context: Any
    is_legacy_evaluation: bool

    @staticmethod
    def create(
        asset_key: AssetKey,
        asset_graph_view: AssetGraphView,
        logger: logging.Logger,
        current_tick_results_by_key: Mapping[AssetKey, AutomationResult],
        condition_cursor: Optional[AutomationConditionCursor],
        legacy_context: "LegacyRuleEvaluationContext",
    ) -> "AutomationContext":
        asset_graph = asset_graph_view.asset_graph
        auto_materialize_policy = check.not_none(asset_graph.get(asset_key).auto_materialize_policy)
        automation_condition = auto_materialize_policy.to_automation_condition()

        return AutomationContext(
            candidate_slice=asset_graph_view.get_asset_slice(asset_key=asset_key),
            condition=automation_condition,
            condition_unique_id=automation_condition.get_unique_id(
                parent_unique_id=None, index=None
            ),
            asset_graph_view=asset_graph_view,
            parent_context=None,
            create_time=pendulum.now("UTC"),
            logger=logger,
            cursor=condition_cursor,
            current_tick_results_by_key=current_tick_results_by_key,
            inner_legacy_context=legacy_context,
            non_agv_instance_interface=NonAGVInstanceInterface(
                asset_graph_view.get_inner_queryer_for_back_compat()
            ),
            is_legacy_evaluation=_has_legacy_condition(automation_condition),
        )

    def for_child_condition(
        self, child_condition: AutomationCondition, child_index: int, candidate_slice: AssetSlice
    ) -> "AutomationContext":
        return AutomationContext(
            candidate_slice=candidate_slice,
            condition=child_condition,
            condition_unique_id=child_condition.get_unique_id(
                parent_unique_id=self.condition_unique_id, index=child_index
            ),
            asset_graph_view=self.asset_graph_view,
            parent_context=self,
            create_time=pendulum.now("UTC"),
            logger=self.logger,
            cursor=self.cursor,
            current_tick_results_by_key=self.current_tick_results_by_key,
            inner_legacy_context=self.inner_legacy_context.for_child(
                child_condition,
                child_condition.get_unique_id(
                    parent_unique_id=self.condition_unique_id, index=child_index
                ),
                candidate_slice.convert_to_valid_asset_subset(),
            ),
            non_agv_instance_interface=self.non_agv_instance_interface,
            is_legacy_evaluation=self.is_legacy_evaluation,
        )

    @property
    def asset_graph(self) -> "BaseAssetGraph":
        return self.asset_graph_view.asset_graph

    @property
    def asset_key(self) -> AssetKey:
        """The asset key over which this condition is being evaluated."""
        return self.candidate_slice.asset_key

    @property
    def partitions_def(self) -> Optional[PartitionsDefinition]:
        """The partitions definition for the asset being evaluated, if it exists."""
        return self.asset_graph.get(self.asset_key).partitions_def

    @property
    def root_context(self) -> "AutomationContext":
        """Returns the context object at the root of the condition evaluation tree."""
        return self.parent_context.root_context if self.parent_context is not None else self

    @property
    def node_cursor(self) -> Optional[AutomationConditionNodeCursor]:
        """Returns the evaluation node for this node from the previous evaluation, if this node
        was evaluated on the previous tick.
        """
        if self.cursor is None:
            return None
        else:
            return self.cursor.node_cursors_by_unique_id.get(self.condition_unique_id)

    @property
    def previous_true_slice(self) -> Optional[AssetSlice]:
        """Returns the true slice for this node from the previous evaluation, if this node was
        evaluated on the previous tick.
        """
        if self.node_cursor is None:
            return None
        else:
            return self.asset_graph_view.get_asset_slice_from_subset(self.node_cursor.true_subset)

    @property
    def effective_dt(self) -> datetime.datetime:
        return self.asset_graph_view.effective_dt

    @property
    def legacy_context(self) -> LegacyRuleEvaluationContext:
        return (
            self.inner_legacy_context
            if self.is_legacy_evaluation
            else check.failed(
                "Legacy access only allowed in AutoMaterializeRule subclasses in auto_materialize_rules_impls.py"
            )
        )

    @property
    def previous_requested_slice(self) -> Optional[AssetSlice]:
        """Returns the requested slice for the previous evaluation. If this asset has never been
        evaluated, returns None.
        """
        return (
            self.asset_graph_view.get_asset_slice_from_subset(self.cursor.previous_requested_subset)
            if self.cursor
            else None
        )

    @property
    def previous_candidate_slice(self) -> Optional[AssetSlice]:
        """Returns the candidate slice for the previous evaluation. If this node has never been
        evaluated, returns None.
        """
        candidate_subset = self.node_cursor.candidate_subset if self.node_cursor else None
        if isinstance(candidate_subset, HistoricalAllPartitionsSubsetSentinel):
            return self.asset_graph_view.get_asset_slice(asset_key=self.asset_key)
        else:
            return (
                self.asset_graph_view.get_asset_slice_from_subset(candidate_subset)
                if candidate_subset
                else None
            )

    @property
    def previous_evaluation_max_storage_id(self) -> Optional[int]:
        """Returns the maximum storage ID for the previous time this asset was evaluated."""
        return self.cursor.temporal_context.last_event_id if self.cursor else None

    @property
    def previous_evaluation_effective_dt(self) -> Optional[datetime.datetime]:
        """Returns the datetime for the previous time this asset was evaluated."""
        return self.cursor.temporal_context.effective_dt if self.cursor else None

    @property
    def new_max_storage_id(self) -> Optional[int]:
        if self.is_legacy_evaluation:
            # legacy evaluations handle event log tailing in a different manner, and so need to
            # use a different storage id cursoring scheme
            return self.legacy_context.new_max_storage_id
        else:
            return self.asset_graph_view.last_event_id

    @property
    def new_temporal_context(self) -> TemporalContext:
        return TemporalContext(
            effective_dt=self.effective_dt, last_event_id=self.new_max_storage_id
        )



from abc import abstractmethod
from typing import AbstractSet, Optional

from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.base_asset_graph import BaseAssetGraph
from dagster._serdes.serdes import whitelist_for_serdes

from ..automation_condition import AutomationCondition, AutomationResult
from ..automation_context import AutomationContext


class DepConditionWrapperCondition(AutomationCondition):
    """Wrapper object which evaluates a condition against a dependency and returns a subset
    representing the subset of downstream asset which has at least one parent which evaluated to
    True.
    """

    dep_key: AssetKey
    operand: AutomationCondition

    @property
    def description(self) -> str:
        return f"{self.dep_key.to_user_string()}"

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        # only evaluate parents of the current candidates
        dep_candidate_slice = context.candidate_slice.compute_parent_slice(self.dep_key)
        dep_context = context.for_child_condition(
            child_condition=self.operand, child_index=0, candidate_slice=dep_candidate_slice
        )

        # evaluate condition against the dependency
        dep_result = self.operand.evaluate(dep_context)

        # find all children of the true dep slice
        true_slice = dep_result.true_slice.compute_child_slice(context.asset_key)
        return AutomationResult.create_from_children(
            context=context, true_slice=true_slice, child_results=[dep_result]
        )


class DepCondition(AutomationCondition):
    operand: AutomationCondition
    allow_selection: Optional[AssetSelection] = None
    ignore_selection: Optional[AssetSelection] = None

    @property
    @abstractmethod
    def base_description(self) -> str: ...

    @property
    def description(self) -> str:
        description = f"{self.base_description} deps"
        if self.allow_selection is not None:
            description += f" within selection {self.allow_selection}"
        if self.ignore_selection is not None:
            description += f" except for {self.ignore_selection}"
        return description

    def allow(self, selection: AssetSelection) -> "DepCondition":
        """Returns a copy of this condition that will only consider dependencies within the provided
        AssetSelection.
        """
        allow_selection = (
            selection if self.allow_selection is None else selection | self.allow_selection
        )
        return self.model_copy(update={"allow_selection": allow_selection})

    def ignore(self, selection: AssetSelection) -> "DepCondition":
        """Returns a copy of this condition that will ignore dependencies within the provided
        AssetSelection.
        """
        ignore_selection = (
            selection if self.ignore_selection is None else selection | self.ignore_selection
        )
        return self.model_copy(update={"ignore_selection": ignore_selection})

    def _get_dep_keys(
        self, asset_key: AssetKey, asset_graph: BaseAssetGraph
    ) -> AbstractSet[AssetKey]:
        dep_keys = asset_graph.get(asset_key).parent_keys
        if self.allow_selection is not None:
            dep_keys &= self.allow_selection.resolve(asset_graph)
        if self.ignore_selection is not None:
            dep_keys -= self.ignore_selection.resolve(asset_graph)
        return dep_keys


@whitelist_for_serdes
class AnyDepsCondition(DepCondition):
    @property
    def base_description(self) -> str:
        return "Any"

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        dep_results = []
        true_slice = context.asset_graph_view.create_empty_slice(asset_key=context.asset_key)

        for i, dep_key in enumerate(
            sorted(self._get_dep_keys(context.asset_key, context.asset_graph))
        ):
            dep_condition = DepConditionWrapperCondition(dep_key=dep_key, operand=self.operand)
            dep_result = dep_condition.evaluate(
                context.for_child_condition(
                    child_condition=dep_condition,
                    child_index=i,
                    candidate_slice=context.candidate_slice,
                )
            )
            dep_results.append(dep_result)
            true_slice = true_slice.compute_union(dep_result.true_slice)

        true_slice = context.candidate_slice.compute_intersection(true_slice)
        return AutomationResult.create_from_children(
            context, true_slice=true_slice, child_results=dep_results
        )


@whitelist_for_serdes
class AllDepsCondition(DepCondition):
    @property
    def base_description(self) -> str:
        return "All"

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        dep_results = []
        true_slice = context.candidate_slice

        for i, dep_key in enumerate(
            sorted(self._get_dep_keys(context.asset_key, context.asset_graph))
        ):
            dep_condition = DepConditionWrapperCondition(dep_key=dep_key, operand=self.operand)
            dep_result = dep_condition.evaluate(
                context.for_child_condition(
                    child_condition=dep_condition,
                    child_index=i,
                    candidate_slice=context.candidate_slice,
                )
            )
            dep_results.append(dep_result)
            true_slice = true_slice.compute_intersection(dep_result.true_slice)

        return AutomationResult.create_from_children(
            context, true_slice=true_slice, child_results=dep_results
        )



import logging
import sys
from contextlib import contextmanager
from typing import Iterable, Mapping, Optional, Sequence, cast

from dagster._core.definitions.instigation_logger import InstigationLogger
from dagster._core.execution.asset_backfill import execute_asset_backfill_iteration
from dagster._core.execution.backfill import BulkActionStatus, PartitionBackfill
from dagster._core.execution.job_backfill import execute_job_backfill_iteration
from dagster._core.workspace.context import IWorkspaceProcessContext
from dagster._daemon.utils import DaemonErrorCapture
from dagster._time import get_current_datetime
from dagster._utils.error import SerializableErrorInfo


@contextmanager
def _get_instigation_logger_if_log_storage_enabled(
    instance, backfill: PartitionBackfill, default_logger: logging.Logger
):
    if instance.backfill_log_storage_enabled():
        evaluation_time = get_current_datetime()
        log_key = [*backfill.log_storage_prefix, evaluation_time.strftime("%Y%m%d_%H%M%S")]
        with InstigationLogger(
            log_key,
            instance,
            repository_name=None,
            instigator_name=backfill.backfill_id,
            logger_name=default_logger.name,
            console_logger=default_logger,
        ) as _logger:
            backfill_logger = cast(logging.Logger, _logger)
            yield backfill_logger

    else:
        yield default_logger


def execute_backfill_iteration(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    debug_crash_flags: Optional[Mapping[str, int]] = None,
) -> Iterable[Optional[SerializableErrorInfo]]:
    instance = workspace_process_context.instance

    in_progress_backfills = instance.get_backfills(status=BulkActionStatus.REQUESTED)
    canceling_backfills = instance.get_backfills(status=BulkActionStatus.CANCELING)

    if not in_progress_backfills and not canceling_backfills:
        logger.debug("No backfill jobs in progress or canceling.")
        yield None
        return

    backfill_jobs = [*in_progress_backfills, *canceling_backfills]

    yield from execute_backfill_jobs(
        workspace_process_context, logger, backfill_jobs, debug_crash_flags
    )


def execute_backfill_jobs(
    workspace_process_context: IWorkspaceProcessContext,
    logger: logging.Logger,
    backfill_jobs: Sequence[PartitionBackfill],
    debug_crash_flags: Optional[Mapping[str, int]] = None,
) -> Iterable[Optional[SerializableErrorInfo]]:
    instance = workspace_process_context.instance

    for backfill_job in backfill_jobs:
        backfill_id = backfill_job.backfill_id

        # refetch, in case the backfill was updated in the meantime
        backfill = cast(PartitionBackfill, instance.get_backfill(backfill_id))
        with _get_instigation_logger_if_log_storage_enabled(instance, backfill, logger) as _logger:
            # create a logger that will always include the backfill_id as an `extra`
            backfill_logger = cast(
                logging.Logger,
                logging.LoggerAdapter(_logger, extra={"backfill_id": backfill.backfill_id}),
            )

            try:
                if backfill.is_asset_backfill:
                    yield from execute_asset_backfill_iteration(
                        backfill, backfill_logger, workspace_process_context, instance
                    )
                else:
                    yield from execute_job_backfill_iteration(
                        backfill,
                        backfill_logger,
                        workspace_process_context,
                        debug_crash_flags,
                        instance,
                    )
            except Exception:
                error_info = DaemonErrorCapture.on_exception(
                    sys.exc_info(),
                    logger=backfill_logger,
                    log_message=f"Backfill failed for {backfill.backfill_id}",
                )
                instance.update_backfill(
                    backfill.with_status(BulkActionStatus.FAILED).with_error(error_info)
                )
                yield error_info



from typing import Optional, Sequence

from dagster._core.asset_graph_view.asset_graph_view import AssetSlice
from dagster._core.definitions.asset_subset import AssetSubset
from dagster._serdes.serdes import whitelist_for_serdes

from ..automation_condition import AutomationCondition, AutomationResult
from ..automation_context import AutomationContext


@whitelist_for_serdes
class NewlyTrueCondition(AutomationCondition):
    operand: AutomationCondition

    @property
    def requires_cursor(self) -> bool:
        return True

    @property
    def description(self) -> str:
        return "Condition newly became true."

    @property
    def children(self) -> Sequence[AutomationCondition]:
        return [self.operand]

    def _get_previous_child_true_slice(self, context: AutomationContext) -> Optional[AssetSlice]:
        """Returns the true slice of the child from the previous tick, which is stored in the
        extra state field of the cursor.
        """
        if not context.node_cursor:
            return None
        true_subset = context.node_cursor.get_extra_state(as_type=AssetSubset)
        if not true_subset:
            return None
        return context.asset_graph_view.get_asset_slice_from_subset(true_subset)

    def evaluate(self, context: AutomationContext) -> AutomationResult:
        # evaluate child condition
        child_context = context.for_child_condition(
            self.operand,
            child_index=0,
            # must evaluate child condition over the entire slice to avoid missing state transitions
            candidate_slice=context.asset_graph_view.get_asset_slice(asset_key=context.asset_key),
        )
        child_result = self.operand.evaluate(child_context)

        # get the set of asset partitions of the child which newly became true
        newly_true_child_slice = child_result.true_slice.compute_difference(
            self._get_previous_child_true_slice(context)
            or context.asset_graph_view.create_empty_slice(asset_key=context.asset_key)
        )

        return AutomationResult.create_from_children(
            context=context,
            true_slice=context.candidate_slice.compute_intersection(newly_true_child_slice),
            child_results=[child_result],
            extra_state=child_result.true_subset,
        )



from typing import Any, Mapping, Optional, cast

import pendulum

import dagster._check as check
from dagster._annotations import experimental
from dagster._core.asset_graph_view.asset_graph_view import AssetGraphView, TemporalContext
from dagster._core.definitions.asset_selection import CoercibleToAssetSelection
from dagster._core.definitions.data_time import CachingDataTimeResolver
from dagster._core.definitions.data_version import CachingStaleStatusResolver
from dagster._core.definitions.declarative_automation.automation_condition_evaluator import (
    AutomationConditionEvaluator,
)
from dagster._core.definitions.run_request import SensorResult
from dagster._utils.caching_instance_queryer import CachingInstanceQueryer

from .asset_selection import AssetSelection
from .sensor_definition import (
    DefaultSensorStatus,
    SensorDefinition,
    SensorEvaluationContext,
    SensorType,
)
from .utils import check_valid_name, normalize_tags


def evaluate_scheduling_conditions(context: SensorEvaluationContext):
    from dagster._core.definitions.asset_daemon_context import build_run_requests
    from dagster._daemon.asset_daemon import (
        asset_daemon_cursor_from_instigator_serialized_cursor,
        asset_daemon_cursor_to_instigator_serialized_cursor,
    )

    asset_graph = check.not_none(context.repository_def).asset_graph

    instance_queryer = CachingInstanceQueryer(
        context.instance,
        asset_graph,
        evaluation_time=pendulum.now(),
        logger=context.log,
    )

    asset_graph_view = AssetGraphView(
        stale_resolver=CachingStaleStatusResolver(
            instance=context.instance,
            asset_graph=asset_graph,
            instance_queryer=instance_queryer,
        ),
        temporal_context=TemporalContext(
            effective_dt=instance_queryer.evaluation_time,
            last_event_id=None,
        ),
    )

    data_time_resolver = CachingDataTimeResolver(
        asset_graph_view.get_inner_queryer_for_back_compat()
    )
    cursor = asset_daemon_cursor_from_instigator_serialized_cursor(
        context.cursor,
        asset_graph,
    )

    evaluator = AutomationConditionEvaluator(
        asset_graph=asset_graph,
        asset_keys=asset_graph.all_asset_keys,
        asset_graph_view=asset_graph_view,
        logger=context.log,
        data_time_resolver=data_time_resolver,
        cursor=cursor,
        respect_materialization_data_versions=True,
        auto_materialize_run_tags={},
    )
    results, to_request = evaluator.evaluate()
    new_cursor = cursor.with_updates(
        evaluation_id=cursor.evaluation_id,
        evaluation_timestamp=instance_queryer.evaluation_time.timestamp(),
        newly_observe_requested_asset_keys=[],  # skip for now, hopefully forever
        condition_cursors=[result.get_new_cursor() for result in results],
    )
    run_requests = build_run_requests(
        asset_partitions=to_request,
        asset_graph=asset_graph,
        # tick_id and sensor tags should get set in daemon
        run_tags=context.instance.auto_materialize_run_tags,
    )

    return SensorResult(
        run_requests=run_requests,
        cursor=asset_daemon_cursor_to_instigator_serialized_cursor(new_cursor),
    )


def not_supported(context):
    raise NotImplementedError(
        "Automation policy sensors cannot be evaluated like regular user-space sensors."
    )


@experimental
class AutoMaterializeSensorDefinition(SensorDefinition):
    """Targets a set of assets and repeatedly evaluates all the AutoMaterializePolicys on all of
    those assets to determine which to request runs for.

    Args:
        name: The name of the sensor.
        asset_selection (Union[str, Sequence[str], Sequence[AssetKey], Sequence[Union[AssetsDefinition, SourceAsset]], AssetSelection]):
            The assets to evaluate AutoMaterializePolicys of and request runs for.
        run_tags: Optional[Mapping[str, Any]] = None,
        default_status (DefaultSensorStatus): Whether the sensor starts as running or not. The default
            status can be overridden from the Dagster UI or via the GraphQL API.
        minimum_interval_seconds (Optional[int]): The frequency at which to try to evaluate the
            sensor. The actual interval will be longer if the sensor evaluation takes longer than
            the provided interval.
        description (Optional[str]): A human-readable description of the sensor.
    """

    def __init__(
        self,
        name: str,
        *,
        asset_selection: CoercibleToAssetSelection,
        run_tags: Optional[Mapping[str, Any]] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        minimum_interval_seconds: Optional[int] = None,
        description: Optional[str] = None,
    ):
        self._run_tags = normalize_tags(run_tags).tags

        super().__init__(
            name=check_valid_name(name),
            job_name=None,
            evaluation_fn=not_supported,
            minimum_interval_seconds=minimum_interval_seconds,
            description=description,
            job=None,
            jobs=None,
            default_status=default_status,
            required_resource_keys=None,
            asset_selection=asset_selection,
        )

    @property
    def run_tags(self) -> Mapping[str, str]:
        return self._run_tags

    @property
    def asset_selection(self) -> AssetSelection:
        return cast(AssetSelection, super().asset_selection)

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.AUTO_MATERIALIZE


@experimental
class AutomationSensorDefinition(SensorDefinition):
    def __init__(
        self,
        name: str,
        *,
        asset_selection: CoercibleToAssetSelection,
        run_tags: Optional[Mapping[str, Any]] = None,
        default_status: DefaultSensorStatus = DefaultSensorStatus.STOPPED,
        minimum_interval_seconds: Optional[int] = None,
    ):
        """Variant of AutoMaterializeSensorDefinition that evaluates automation conditions in
        user code.
        """
        self._run_tags = normalize_tags(run_tags).tags

        super().__init__(
            name=check_valid_name(name),
            job_name=None,
            evaluation_fn=evaluate_scheduling_conditions,
            minimum_interval_seconds=minimum_interval_seconds,
            default_status=default_status,
            asset_selection=asset_selection,
        )

    @property
    def run_tags(self) -> Mapping[str, str]:
        return self._run_tags

    @property
    def asset_selection(self) -> AssetSelection:
        return cast(AssetSelection, super().asset_selection)

    @property
    def sensor_type(self) -> SensorType:
        return SensorType.AUTOMATION



from abc import ABC, abstractmethod
from collections import deque
from datetime import datetime
from functools import cached_property, total_ordering
from heapq import heapify, heappop, heappush
from typing import (
    TYPE_CHECKING,
    AbstractSet,
    Callable,
    Dict,
    Generic,
    Iterable,
    Iterator,
    List,
    Mapping,
    NamedTuple,
    Optional,
    Sequence,
    Set,
    Tuple,
    TypeVar,
    Union,
    cast,
)

import dagster._check as check
from dagster._core.definitions.asset_check_spec import AssetCheckKey
from dagster._core.definitions.asset_subset import ValidAssetSubset
from dagster._core.definitions.backfill_policy import BackfillPolicy
from dagster._core.definitions.events import AssetKey
from dagster._core.definitions.freshness_policy import FreshnessPolicy
from dagster._core.definitions.metadata import ArbitraryMetadataMapping
from dagster._core.definitions.partition import PartitionsDefinition
from dagster._core.definitions.partition_mapping import PartitionMapping
from dagster._core.errors import DagsterInvalidInvocationError
from dagster._core.instance import DynamicPartitionsStore
from dagster._core.selector.subset_selector import DependencyGraph, fetch_sources
from dagster._core.utils import toposort
from dagster._utils.cached_method import cached_method

from .events import AssetKeyPartitionKey
from .partition import PartitionsSubset
from .partition_key_range import PartitionKeyRange
from .partition_mapping import UpstreamPartitionsResult, infer_partition_mapping
from .time_window_partitions import get_time_partition_key, get_time_partitions_def

if TYPE_CHECKING:
    from dagster._core.definitions.asset_graph_subset import AssetGraphSubset
    from dagster._core.definitions.auto_materialize_policy import AutoMaterializePolicy

AssetKeyOrCheckKey = Union[AssetKey, AssetCheckKey]


class ParentsPartitionsResult(NamedTuple):
    """Represents the result of mapping an asset partition to its upstream parent partitions.

    parent_partitions (AbstractSet[AssetKeyPartitionKey]): The existent parent partitions that are
        upstream of the asset partition. Filters out nonexistent partitions.
    required_but_nonexistent_parents_partitions (AbstractSet[AssetKeyPartitionKey]): The required
        upstream asset partitions that were mapped to but do not exist.
    """

    parent_partitions: AbstractSet[AssetKeyPartitionKey]
    required_but_nonexistent_parents_partitions: AbstractSet[AssetKeyPartitionKey]


class BaseAssetNode(ABC):
    key: AssetKey
    parent_keys: AbstractSet[AssetKey]
    child_keys: AbstractSet[AssetKey]

    @property
    def has_self_dependency(self) -> bool:
        return self.key in self.parent_keys

    @property
    @abstractmethod
    def description(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def group_name(self) -> str: ...

    @property
    @abstractmethod
    def is_materializable(self) -> bool: ...

    @property
    @abstractmethod
    def is_observable(self) -> bool: ...

    @property
    @abstractmethod
    def is_external(self) -> bool: ...

    @property
    @abstractmethod
    def is_executable(self) -> bool: ...

    @property
    @abstractmethod
    def metadata(self) -> ArbitraryMetadataMapping: ...

    @property
    @abstractmethod
    def tags(self) -> Mapping[str, str]: ...

    @property
    @abstractmethod
    def owners(self) -> Sequence[str]: ...

    @property
    @abstractmethod
    def is_partitioned(self) -> bool: ...

    @property
    @abstractmethod
    def partitions_def(self) -> Optional[PartitionsDefinition]: ...

    @property
    @abstractmethod
    def partition_mappings(self) -> Mapping[AssetKey, PartitionMapping]: ...

    @property
    @abstractmethod
    def freshness_policy(self) -> Optional[FreshnessPolicy]: ...

    @property
    @abstractmethod
    def auto_materialize_policy(self) -> Optional["AutoMaterializePolicy"]: ...

    @property
    @abstractmethod
    def auto_observe_interval_minutes(self) -> Optional[float]: ...

    @property
    @abstractmethod
    def backfill_policy(self) -> Optional[BackfillPolicy]: ...

    @property
    @abstractmethod
    def code_version(self) -> Optional[str]: ...

    @property
    @abstractmethod
    def check_keys(self) -> AbstractSet[AssetCheckKey]: ...

    @property
    @abstractmethod
    def execution_set_asset_keys(self) -> AbstractSet[AssetKey]: ...

    @property
    @abstractmethod
    def execution_set_asset_and_check_keys(
        self,
    ) -> AbstractSet[Union[AssetKey, AssetCheckKey]]: ...

    def __str__(self) -> str:
        return f"{self.__class__.__name__}<{self.key.to_user_string()}>"


T_AssetNode = TypeVar("T_AssetNode", bound=BaseAssetNode)


class BaseAssetGraph(ABC, Generic[T_AssetNode]):
    _asset_nodes_by_key: Mapping[AssetKey, T_AssetNode]

    @property
    def asset_nodes(self) -> Iterable[T_AssetNode]:
        return self._asset_nodes_by_key.values()

    def has(self, asset_key: AssetKey) -> bool:
        return asset_key in self._asset_nodes_by_key

    def get(self, asset_key: AssetKey) -> T_AssetNode:
        return self._asset_nodes_by_key[asset_key]

    @cached_property
    def asset_dep_graph(self) -> DependencyGraph[AssetKey]:
        return {
            "upstream": {node.key: node.parent_keys for node in self.asset_nodes},
            "downstream": {node.key: node.child_keys for node in self.asset_nodes},
        }

    @cached_property
    def all_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes}

    @cached_property
    def materializable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if node.is_materializable}

    @cached_property
    def observable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if node.is_observable}

    @cached_property
    def external_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if node.is_external}

    @cached_property
    def executable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if node.is_executable}

    @cached_property
    def unexecutable_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if not node.is_executable}

    @cached_property
    def toposorted_asset_keys(self) -> Sequence[AssetKey]:
        """Return topologically sorted asset keys in graph. Keys with the same topological level are
        sorted alphabetically to provide stability.
        """
        return [
            item
            for items_in_level in toposort(self.asset_dep_graph["upstream"])
            for item in sorted(items_in_level)
        ]

    @cached_property
    def toposorted_asset_keys_by_level(self) -> Sequence[AbstractSet[AssetKey]]:
        """Return topologically sorted asset keys grouped into sets containing keys of the same
        topological level.
        """
        return [set(level) for level in toposort(self.asset_dep_graph["upstream"])]

    @cached_property
    def unpartitioned_asset_keys(self) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if not node.is_partitioned}

    def asset_keys_for_group(self, group_name: str) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if node.group_name == group_name}

    @cached_method
    def asset_keys_for_partitions_def(
        self, partitions_def: PartitionsDefinition
    ) -> AbstractSet[AssetKey]:
        return {node.key for node in self.asset_nodes if node.partitions_def == partitions_def}

    @cached_property
    def root_materializable_asset_keys(self) -> AbstractSet[AssetKey]:
        """Materializable asset keys that have no materializable parents."""
        from .asset_selection import KeysAssetSelection

        return (
            KeysAssetSelection(selected_keys=list(self.materializable_asset_keys))
            .roots()
            .resolve(self)
        )

    @cached_property
    def root_executable_asset_keys(self) -> AbstractSet[AssetKey]:
        """Executable asset keys that have no executable parents."""
        return fetch_sources(
            self.asset_dep_graph, self.observable_asset_keys | self.materializable_asset_keys
        )

    @property
    @abstractmethod
    def asset_check_keys(self) -> AbstractSet[AssetCheckKey]: ...

    @cached_property
    def orphan_asset_check_keys(self) -> AbstractSet[AssetCheckKey]:
        """Asset check keys that target an asset with no corresponding executable definition in the graph."""
        return {k for k in self.asset_check_keys if k.asset_key not in self.executable_asset_keys}

    @cached_property
    def all_partitions_defs(self) -> Sequence[PartitionsDefinition]:
        return sorted(
            set(node.partitions_def for node in self.asset_nodes if node.partitions_def), key=repr
        )

    @cached_property
    def all_group_names(self) -> AbstractSet[str]:
        return {a.group_name for a in self.asset_nodes if a.group_name is not None}

    def get_partition_mapping(
        self, asset_key: AssetKey, parent_asset_key: AssetKey
    ) -> PartitionMapping:
        node = self.get(asset_key)
        return infer_partition_mapping(
            node.partition_mappings.get(parent_asset_key),
            node.partitions_def,
            self.get(parent_asset_key).partitions_def,
        )

    def get_children(self, node: T_AssetNode) -> AbstractSet[T_AssetNode]:
        """Returns all asset nodes that directly depend on the given asset node."""
        return {self._asset_nodes_by_key[key] for key in self.get(node.key).child_keys}

    def get_parents(self, node: T_AssetNode) -> AbstractSet[T_AssetNode]:
        """Returns all asset nodes that are direct dependencies on the given asset node."""
        return {self._asset_nodes_by_key[key] for key in self.get(node.key).parent_keys}

    def get_ancestor_asset_keys(
        self, asset_key: AssetKey, include_self: bool = False
    ) -> AbstractSet[AssetKey]:
        """Returns all nth-order dependencies of an asset."""
        ancestors = set()
        next_parents = self.get(asset_key).parent_keys - {asset_key}  # remove self-dependencies
        while next_parents:
            pending_next_parents = set()
            for node_key in next_parents:
                if node_key in ancestors:
                    continue
                ancestors.add(node_key)
                pending_next_parents.update(self.get(node_key).parent_keys)

            next_parents = pending_next_parents

        if include_self:
            ancestors.add(asset_key)
        return ancestors

    def get_partitions_in_range(
        self,
        asset_key: AssetKey,
        partition_key_range: PartitionKeyRange,
        dynamic_partitions_store: DynamicPartitionsStore,
    ) -> Sequence[AssetKeyPartitionKey]:
        partition_def = self.get(asset_key).partitions_def
        partition_keys_in_range = check.not_none(partition_def).get_partition_keys_in_range(
            partition_key_range, dynamic_partitions_store
        )
        return [
            AssetKeyPartitionKey(asset_key, partition_key)
            for partition_key in partition_keys_in_range
        ]

    def get_parent_asset_subset(
        self,
        child_asset_subset: ValidAssetSubset,
        parent_asset_key: AssetKey,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
    ) -> ValidAssetSubset:
        """Given a child AssetSubset, returns the corresponding parent AssetSubset, based on the
        relevant PartitionMapping.
        """
        child_asset_key = child_asset_subset.asset_key
        child_partitions_def = self.get(child_asset_key).partitions_def
        parent_partitions_def = self.get(parent_asset_key).partitions_def

        if parent_partitions_def is None:
            return ValidAssetSubset(asset_key=parent_asset_key, value=child_asset_subset.size > 0)

        partition_mapping = self.get_partition_mapping(child_asset_key, parent_asset_key)
        parent_partitions_subset = (
            partition_mapping.get_upstream_mapped_partitions_result_for_partitions(
                child_asset_subset.subset_value if child_partitions_def is not None else None,
                downstream_partitions_def=child_partitions_def,
                upstream_partitions_def=parent_partitions_def,
                dynamic_partitions_store=dynamic_partitions_store,
                current_time=current_time,
            )
        ).partitions_subset

        return ValidAssetSubset(asset_key=parent_asset_key, value=parent_partitions_subset)

    def get_child_asset_subset(
        self,
        parent_asset_subset: ValidAssetSubset,
        child_asset_key: AssetKey,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
    ) -> ValidAssetSubset:
        """Given a parent AssetSubset, returns the corresponding child AssetSubset, based on the
        relevant PartitionMapping.
        """
        parent_asset_key = parent_asset_subset.asset_key
        parent_partitions_def = self.get(parent_asset_key).partitions_def
        child_partitions_def = self.get(child_asset_key).partitions_def

        if parent_partitions_def is None:
            if parent_asset_subset.size > 0:
                return ValidAssetSubset.all(
                    child_asset_key, child_partitions_def, dynamic_partitions_store, current_time
                )
            else:
                return ValidAssetSubset.empty(child_asset_key, child_partitions_def)

        if child_partitions_def is None:
            return ValidAssetSubset(asset_key=child_asset_key, value=parent_asset_subset.size > 0)
        else:
            partition_mapping = self.get_partition_mapping(child_asset_key, parent_asset_key)
            child_partitions_subset = partition_mapping.get_downstream_partitions_for_partitions(
                parent_asset_subset.subset_value,
                parent_partitions_def,
                downstream_partitions_def=child_partitions_def,
                dynamic_partitions_store=dynamic_partitions_store,
                current_time=current_time,
            )
            return ValidAssetSubset(asset_key=child_asset_key, value=child_partitions_subset)

    def get_children_partitions(
        self,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
        asset_key: AssetKey,
        partition_key: Optional[str] = None,
    ) -> AbstractSet[AssetKeyPartitionKey]:
        """Returns every partition in every of the given asset's children that depends on the given
        partition of that asset.
        """
        result: Set[AssetKeyPartitionKey] = set()
        for child in self.get_children(self.get(asset_key)):
            if child.is_partitioned:
                for child_partition_key in self.get_child_partition_keys_of_parent(
                    dynamic_partitions_store,
                    partition_key,
                    asset_key,
                    child.key,
                    current_time,
                ):
                    result.add(AssetKeyPartitionKey(child.key, child_partition_key))
            else:
                result.add(AssetKeyPartitionKey(child.key))
        return result

    def get_child_partition_keys_of_parent(
        self,
        dynamic_partitions_store: DynamicPartitionsStore,
        parent_partition_key: Optional[str],
        parent_asset_key: AssetKey,
        child_asset_key: AssetKey,
        current_time: datetime,
    ) -> Sequence[str]:
        """Converts a partition key from one asset to the corresponding partition keys in a downstream
        asset. Uses the existing partition mapping between the child asset and the parent asset.

        Args:
            parent_partition_key (Optional[str]): The partition key to convert.
            parent_asset_key (AssetKey): The asset key of the upstream asset, which the provided
                partition key belongs to.
            child_asset_key (AssetKey): The asset key of the downstream asset. The provided partition
                key will be mapped to partitions within this asset.

        Returns:
            Sequence[str]: A list of the corresponding downstream partitions in child_asset_key that
                partition_key maps to.
        """
        child_partitions_def = self.get(child_asset_key).partitions_def
        parent_partitions_def = self.get(parent_asset_key).partitions_def

        if child_partitions_def is None:
            raise DagsterInvalidInvocationError(
                f"Asset key {child_asset_key} is not partitioned. Cannot get partition keys."
            )
        if parent_partition_key is None:
            return child_partitions_def.get_partition_keys(
                dynamic_partitions_store=dynamic_partitions_store
            )

        if parent_partitions_def is None:
            raise DagsterInvalidInvocationError(
                f"Parent partition key '{parent_partition_key}' provided, but parent asset"
                f" '{parent_asset_key}' is not partitioned."
            )

        partition_mapping = self.get_partition_mapping(child_asset_key, parent_asset_key)
        child_partitions_subset = partition_mapping.get_downstream_partitions_for_partitions(
            parent_partitions_def.empty_subset().with_partition_keys([parent_partition_key]),
            parent_partitions_def,
            downstream_partitions_def=child_partitions_def,
            dynamic_partitions_store=dynamic_partitions_store,
            current_time=current_time,
        )

        return list(child_partitions_subset.get_partition_keys())

    def get_parents_partitions(
        self,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
        asset_key: AssetKey,
        partition_key: Optional[str] = None,
    ) -> ParentsPartitionsResult:
        """Returns every partition in every of the given asset's parents that the given partition of
        that asset depends on.
        """
        valid_parent_partitions: Set[AssetKeyPartitionKey] = set()
        required_but_nonexistent_parent_partitions: Set[AssetKeyPartitionKey] = set()
        for parent_asset_key in self.get(asset_key).parent_keys:
            if self.has(parent_asset_key) and self.get(parent_asset_key).is_partitioned:
                mapped_partitions_result = self.get_parent_partition_keys_for_child(
                    partition_key,
                    parent_asset_key,
                    asset_key,
                    dynamic_partitions_store=dynamic_partitions_store,
                    current_time=current_time,
                )

                valid_parent_partitions.update(
                    {
                        AssetKeyPartitionKey(parent_asset_key, valid_partition)
                        for valid_partition in mapped_partitions_result.partitions_subset.get_partition_keys()
                    }
                )
                required_but_nonexistent_parent_partitions.update(
                    {
                        AssetKeyPartitionKey(parent_asset_key, invalid_partition)
                        for invalid_partition in mapped_partitions_result.required_but_nonexistent_partition_keys
                    }
                )
            else:
                valid_parent_partitions.add(AssetKeyPartitionKey(parent_asset_key))

        return ParentsPartitionsResult(
            valid_parent_partitions, required_but_nonexistent_parent_partitions
        )

    def get_parent_partition_keys_for_child(
        self,
        partition_key: Optional[str],
        parent_asset_key: AssetKey,
        child_asset_key: AssetKey,
        dynamic_partitions_store: DynamicPartitionsStore,
        current_time: datetime,
    ) -> UpstreamPartitionsResult:
        """Converts a partition key from one asset to the corresponding partition keys in one of its
        parent assets. Uses the existing partition mapping between the child asset and the parent
        asset.

        Args:
            partition_key (Optional[str]): The partition key to convert.
            child_asset_key (AssetKey): The asset key of the child asset, which the provided
                partition key belongs to.
            parent_asset_key (AssetKey): The asset key of the parent asset. The provided partition
                key will be mapped to partitions within this asset.

        Returns:
            Sequence[str]: A list of the corresponding downstream partitions in child_asset_key that
                partition_key maps to.
        """
        partition_key = check.opt_str_param(partition_key, "partition_key")

        child_partitions_def = cast(PartitionsDefinition, self.get(child_asset_key).partitions_def)
        parent_partitions_def = self.get(parent_asset_key).partitions_def

        if parent_partitions_def is None:
            raise DagsterInvalidInvocationError(
                f"Asset key {parent_asset_key} is not partitioned. Cannot get partition keys."
            )

        partition_mapping = self.get_partition_mapping(child_asset_key, parent_asset_key)

        return partition_mapping.get_upstream_mapped_partitions_result_for_partitions(
            (
                child_partitions_def.subset_with_partition_keys([partition_key])
                if partition_key
                else None
            ),
            downstream_partitions_def=child_partitions_def,
            upstream_partitions_def=parent_partitions_def,
            dynamic_partitions_store=dynamic_partitions_store,
            current_time=current_time,
        )

    def has_materializable_parents(self, asset_key: AssetKey) -> bool:
        """Determines if an asset has any parents which are materializable."""
        if self.get(asset_key).is_external:
            return False
        return any(
            self.has(parent_key) and self.get(parent_key).is_materializable
            for parent_key in self.get(asset_key).parent_keys - {asset_key}
        )

    def get_materializable_roots(self, asset_key: AssetKey) -> AbstractSet[AssetKey]:
        """Returns all assets upstream of the given asset which do not consume any other
        materializable assets.
        """
        if not self.has_materializable_parents(asset_key):
            return {asset_key}
        return {
            key
            for key in self.upstream_key_iterator(asset_key)
            if self.has(key)
            and self.get(key).is_materializable
            and not self.has_materializable_parents(key)
        }

    def upstream_key_iterator(self, asset_key: AssetKey) -> Iterator[AssetKey]:
        """Iterates through all asset keys which are upstream of the given key."""
        visited: Set[AssetKey] = set()
        queue = deque([asset_key])
        while queue:
            current_key = queue.popleft()
            for parent_key in self.get(current_key).parent_keys:
                if parent_key not in visited:
                    yield parent_key
                    queue.append(parent_key)
                    visited.add(parent_key)

    @abstractmethod
    def get_execution_set_asset_and_check_keys(
        self, asset_key_or_check_key: AssetKeyOrCheckKey
    ) -> AbstractSet[AssetKeyOrCheckKey]:
        """For a given asset/check key, return the set of asset/check keys that must be
        materialized/computed at the same time.
        """
        ...

    @cached_method
    def get_downstream_freshness_policies(
        self, *, asset_key: AssetKey
    ) -> AbstractSet[FreshnessPolicy]:
        asset = self.get(asset_key)
        downstream_policies = set().union(
            *(
                self.get_downstream_freshness_policies(asset_key=child_key)
                for child_key in self.get(asset_key).child_keys
                if child_key != asset_key
            )
        )
        if asset.partitions_def is None and asset.freshness_policy is not None:
            downstream_policies.add(asset.freshness_policy)

        return downstream_policies

    def bfs_filter_subsets(
        self,
        dynamic_partitions_store: DynamicPartitionsStore,
        condition_fn: Callable[[AssetKey, Optional[PartitionsSubset]], bool],
        initial_subset: "AssetGraphSubset",
        current_time: datetime,
    ) -> "AssetGraphSubset":
        """Returns asset partitions within the graph that satisfy supplied criteria.

        - Are >= initial_asset_partitions
        - Asset matches the condition_fn
        - Any of their ancestors >= initial_asset_partitions match the condition_fn.

        Visits parents before children.
        """
        from .asset_graph_subset import AssetGraphSubset

        all_assets = set(initial_subset.asset_keys)
        check.invariant(
            len(initial_subset.asset_keys) == 1, "Multiple initial assets not yet supported"
        )
        initial_asset_key = next(iter(initial_subset.asset_keys))
        queue = deque([initial_asset_key])

        queued_subsets_by_asset_key: Dict[AssetKey, Optional[PartitionsSubset]] = {
            initial_asset_key: (
                initial_subset.get_partitions_subset(initial_asset_key, self)
                if self.get(initial_asset_key).is_partitioned
                else None
            ),
        }
        result = AssetGraphSubset()

        while len(queue) > 0:
            asset_key = queue.popleft()
            partitions_subset = queued_subsets_by_asset_key.get(asset_key)

            if condition_fn(asset_key, partitions_subset):
                result |= AssetGraphSubset(
                    non_partitioned_asset_keys={asset_key} if partitions_subset is None else set(),
                    partitions_subsets_by_asset_key=(
                        {asset_key: partitions_subset} if partitions_subset is not None else {}
                    ),
                )

                for child_key in self.get(asset_key).child_keys:
                    partition_mapping = self.get_partition_mapping(child_key, asset_key)
                    child_partitions_def = self.get(child_key).partitions_def

                    if child_partitions_def:
                        if partitions_subset is None:
                            child_partitions_subset = (
                                child_partitions_def.subset_with_all_partitions(
                                    current_time=current_time,
                                    dynamic_partitions_store=dynamic_partitions_store,
                                )
                            )
                            queued_subsets_by_asset_key[child_key] = child_partitions_subset
                        else:
                            child_partitions_subset = (
                                partition_mapping.get_downstream_partitions_for_partitions(
                                    partitions_subset,
                                    check.not_none(self.get(asset_key).partitions_def),
                                    downstream_partitions_def=child_partitions_def,
                                    dynamic_partitions_store=dynamic_partitions_store,
                                    current_time=current_time,
                                )
                            )
                            prior_child_partitions_subset = queued_subsets_by_asset_key.get(
                                child_key
                            )
                            queued_subsets_by_asset_key[child_key] = (
                                child_partitions_subset
                                if not prior_child_partitions_subset
                                else child_partitions_subset | prior_child_partitions_subset
                            )
                    else:
                        child_partitions_subset = None

                    if child_key not in all_assets:
                        queue.append(child_key)
                        all_assets.add(child_key)

        return result

    def bfs_filter_asset_partitions(
        self,
        dynamic_partitions_store: DynamicPartitionsStore,
        condition_fn: Callable[
            [Iterable[AssetKeyPartitionKey], AbstractSet[AssetKeyPartitionKey]], Tuple[bool, str]
        ],
        initial_asset_partitions: Iterable[AssetKeyPartitionKey],
        evaluation_time: datetime,
    ) -> Tuple[
        AbstractSet[AssetKeyPartitionKey], Sequence[Tuple[Iterable[AssetKeyPartitionKey], str]]
    ]:
        """Returns asset partitions within the graph that satisfy supplied criteria.

        - Are >= initial_asset_partitions
        - Match the condition_fn
        - Any of their ancestors >= initial_asset_partitions match the condition_fn

        Also returns a list of tuples, where each tuple is a candidated_unit (list of
        AssetKeyPartitionKeys that must be materialized together - ie multi_asset) that do not
        satisfy the criteria and the reason they were filtered out.

        The condition_fn should return a tuple of a boolean indicating whether the asset partition meets
        the condition and a string explaining why it does not meet the condition, if applicable.

        Visits parents before children.

        When asset partitions are part of the same execution set (non-subsettable multi-asset),
        they're provided all at once to the condition_fn.
        """
        all_nodes = set(initial_asset_partitions)

        # invariant: we never consider an asset partition before considering its ancestors
        queue = ToposortedPriorityQueue(self, all_nodes, include_full_execution_set=True)

        result: Set[AssetKeyPartitionKey] = set()
        failed_reasons: List[Tuple[Iterable[AssetKeyPartitionKey], str]] = []

        while len(queue) > 0:
            candidates_unit = queue.dequeue()

            meets_condition, fail_reason = condition_fn(candidates_unit, result)
            if meets_condition:
                result.update(candidates_unit)

                for candidate in candidates_unit:
                    for child in self.get_children_partitions(
                        dynamic_partitions_store,
                        evaluation_time,
                        candidate.asset_key,
                        candidate.partition_key,
                    ):
                        if child not in all_nodes:
                            queue.enqueue(child)
                            all_nodes.add(child)
            else:
                failed_reasons.append((candidates_unit, fail_reason))

        return result, failed_reasons

    def split_asset_keys_by_repository(
        self, asset_keys: AbstractSet[AssetKey]
    ) -> Sequence[AbstractSet[AssetKey]]:
        return [asset_keys]

    def __hash__(self) -> int:
        return id(self)

    def __eq__(self, other: object) -> bool:
        return self is other


def sort_key_for_asset_partition(
    asset_graph: BaseAssetGraph, asset_partition: AssetKeyPartitionKey
) -> float:
    """Returns an integer sort key such that asset partitions are sorted in the order in which they
    should be materialized. For assets without a time window partition dimension, this is always 0.
    Assets with a time window partition dimension will be sorted from newest to oldest, unless they
    have a self-dependency, in which case they are sorted from oldest to newest.
    """
    partitions_def = asset_graph.get(asset_partition.asset_key).partitions_def
    time_partitions_def = get_time_partitions_def(partitions_def)
    if time_partitions_def is None:
        return 0

    # A sort key such that time window partitions are sorted from oldest to newest
    time_partition_key = get_time_partition_key(partitions_def, asset_partition.partition_key)
    partition_timestamp = time_partitions_def.start_time_for_partition_key(
        time_partition_key
    ).timestamp()

    if asset_graph.get(asset_partition.asset_key).has_self_dependency:
        # sort self dependencies from oldest to newest, as older partitions must exist before
        # new ones can execute
        return partition_timestamp
    else:
        # sort non-self dependencies from newest to oldest, as newer partitions are more relevant
        # than older ones
        return -1 * partition_timestamp


class ToposortedPriorityQueue:
    """Queue that returns parents before their children."""

    @total_ordering
    class QueueItem(NamedTuple):
        level: int
        partition_sort_key: Optional[float]
        asset_partitions: Iterable[AssetKeyPartitionKey]

        def __eq__(self, other: object) -> bool:
            if isinstance(other, ToposortedPriorityQueue.QueueItem):
                return (
                    self.level == other.level
                    and self.partition_sort_key == other.partition_sort_key
                )
            return False

        def __lt__(self, other: object) -> bool:
            if isinstance(other, ToposortedPriorityQueue.QueueItem):
                return self.level < other.level or (
                    self.level == other.level
                    and self.partition_sort_key is not None
                    and other.partition_sort_key is not None
                    and self.partition_sort_key < other.partition_sort_key
                )
            raise TypeError()

    def __init__(
        self,
        asset_graph: BaseAssetGraph,
        items: Iterable[AssetKeyPartitionKey],
        include_full_execution_set: bool,
    ):
        self._asset_graph = asset_graph
        self._include_full_execution_set = include_full_execution_set

        self._toposort_level_by_asset_key = {
            asset_key: i
            for i, asset_keys in enumerate(asset_graph.toposorted_asset_keys_by_level)
            for asset_key in asset_keys
        }
        self._heap = [self._queue_item(asset_partition) for asset_partition in items]
        heapify(self._heap)

    def enqueue(self, asset_partition: AssetKeyPartitionKey) -> None:
        heappush(self._heap, self._queue_item(asset_partition))

    def dequeue(self) -> Iterable[AssetKeyPartitionKey]:
        # For multi-assets, will include all required multi-asset keys if
        # include_full_execution_set is set to True, or a list of size 1 with just the passed in
        # asset key if it was not.
        return heappop(self._heap).asset_partitions

    def _queue_item(
        self, asset_partition: AssetKeyPartitionKey
    ) -> "ToposortedPriorityQueue.QueueItem":
        asset_key = asset_partition.asset_key

        if self._include_full_execution_set:
            execution_set_keys = self._asset_graph.get(asset_key).execution_set_asset_keys
        else:
            execution_set_keys = {asset_key}

        level = max(
            self._toposort_level_by_asset_key[asset_key] for asset_key in execution_set_keys
        )

        return ToposortedPriorityQueue.QueueItem(
            level,
            sort_key_for_asset_partition(self._asset_graph, asset_partition),
            [AssetKeyPartitionKey(ak, asset_partition.partition_key) for ak in execution_set_keys],
        )

    def __len__(self) -> int:
        return len(self._heap)



from .automation_condition import AutomationCondition as AutomationCondition
from .automation_condition_tester import (
    evaluate_automation_conditions as evaluate_automation_conditions,
)
from .legacy import RuleCondition as RuleCondition
from .legacy.asset_condition import AssetCondition as AssetCondition
from .operands import (
    CronTickPassedCondition as CronTickPassedCondition,
    FailedAutomationCondition as FailedAutomationCondition,
    InLatestTimeWindowCondition as InLatestTimeWindowCondition,
    InProgressAutomationCondition as InProgressAutomationCondition,
    MissingAutomationCondition as MissingAutomationCondition,
    NewlyRequestedCondition as NewlyRequestedCondition,
    NewlyUpdatedCondition as NewlyUpdatedCondition,
    WillBeRequestedCondition as WillBeRequestedCondition,
)
from .operators import (
    AllDepsCondition as AllDepsCondition,
    AndAssetCondition as AndAssetCondition,
    AnyDepsCondition as AnyDepsCondition,
    NewlyTrueCondition as NewlyTrueCondition,
    NotAssetCondition as NotAssetCondition,
    OrAssetCondition as OrAssetCondition,
    SinceCondition as SinceCondition,
)
from .serialized_objects import (
    AssetConditionEvaluationState as AssetConditionEvaluationState,
    AssetConditionSnapshot as AssetConditionSnapshot,
    AssetSubsetWithMetadata as AssetSubsetWithMetadata,
    HistoricalAllPartitionsSubsetSentinel as HistoricalAllPartitionsSubsetSentinel,
)
